//#####################################################################
// Copyright 2022, Jason Goulding, Mehrnaz Ayazi
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CAHN_HILLIARD__
#define __CAHN_HILLIARD__
#include <Core/Advection/ADVECTION_CONSERVATIVE_ENO.h>
#include <Core/Advection/ADVECTION_CONSERVATIVE_WENO.h>
#include <Core/Arrays/ARRAYS_ND.h>
#include <Core/Arrays/ARRAYS_ND_BASE.h>
#include <Core/Boundaries/BOUNDARY.h>
#include <Core/Boundaries/BOUNDARY_PERIODIC.h>
#include <Core/Data_Structures/HASHTABLE.h>
#include <Core/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Poisson/PROJECTION_UNIFORM.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Core/Vectors/VECTOR.h>
#include <Core/Viewer/VIEWER_OUTPUT.h>
#include <functional>

namespace PhysBAM{

template<class TV> class CH_SYMBOLIC;

template<class TV>
class CAHN_HILLIARD
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<int,TV::m> IV;

protected:

    T mass_scale=1;
    T time_scale=1;
    T distance_scale=1;
    int ghost=1; // number of cells in the ghost region (used for flux)
    T order=1;

    bool print_gamma_stats=true;
    bool print_status=true;
    bool print_solve_results=true;
    bool print_solve_residual=false;
    bool debug_output=false;
 
    bool suppress_output=false;
    bool error_output=false;
    bool relative_error=false;
    VIEWER_OUTPUT* viewer_output=nullptr; // pointer allocated outside class
    DEBUG_PARTICLES<TV>* debug_particles=nullptr; // pointer allocated outside class
    HASHTABLE<std::string,DEBUG_PARTICLES<TV>*>* error_particles=nullptr; // pointer allocated outside class

    const ARRAY<bool,FACE_INDEX<TV::m> >* psi_N=nullptr; // pointer allocated outside class
    const ARRAY<T,FACE_INDEX<TV::m> >* psi_N_values=nullptr; // pointer allocated outside class

    const ARRAYS_ND_BASE<bool,IV>* psi_D=nullptr; // pointer allocated outside class
    const ARRAYS_ND_BASE<T,IV>* psi_D_values=nullptr; // pointer allocated outside class

    const GRID<TV>& grid;
    CH_SYMBOLIC<TV>* symbolic=nullptr; // pointer allocated inside class
    BOUNDARY_PERIODIC<TV,T> periodic_boundary_scalar;
    ADVECTION_CONSERVATIVE_WENO<TV,T> advection_scalar;
    VECTOR<bool,TV::m> periodic;
    
    void Gradient(ARRAY<T,FACE_INDEX<TV::m> >& out,const ARRAYS_ND_BASE<T,IV>& in_ghost) const;
    void Laplacian(ARRAY<T,IV>& out,const ARRAYS_ND_BASE<T,IV>& in_ghost,int ghost=0) const;
    void Fill_Ghost(ARRAYS_ND_BASE<T,IV>& output,const ARRAYS_ND_BASE<T,IV>& input,int ghost=3,std::function<T(const IV&,int)> boundary_N=0) const;
    void Discrete_Advection(const GRID<TV>& grid,ARRAY<T,IV>& out,const ARRAY<T,IV>& Z,const ARRAY<T,FACE_INDEX<TV::m> >& u,T dt,T time,ARRAY<T,FACE_INDEX<TV::m> >* flux);
    void Advect(ARRAY<T,IV>& output,const ARRAY<T,IV>& g_ghost,const ARRAY<T,FACE_INDEX<TV::m> >& u_ghost,T dt,T time);
    void Print_Gamma_Stats(const ARRAYS_ND_BASE<T,IV>& gamma,std::string title) const;
    void Test_Analytic(const ARRAYS_ND_BASE<T,IV>& data,const char* name,T t,int g=0,bool relative=false);
    void Test_Flux(const GRID<TV>& flux_grid,const ARRAYS_ND_BASE<T,IV>& base,const ARRAYS_ND_BASE<T,IV>& control,const ARRAY<T,FACE_INDEX<TV::m> >& flux,T scale,const std::string title) const;
    
public:
    
    T eta=(T).1; // interface width
    T sigma=1; // surface tension coefficient
    T M=(T)1e-3; // mobility constant
    T lambda=1;

    bool linearize=false;
    bool test_units=false;
    bool test_flux=false;
    bool discrete_advection=false;

    VECTOR<T,TV::m*2> side_contact_angle; // -x, +x, -y, +y, -z, +z

    ARRAY<T,FACE_INDEX<TV::m> > advect_flux;
    ARRAY<T,FACE_INDEX<TV::m> > flux;
   
    CAHN_HILLIARD(const GRID<TV>& g,int ghost,PARSE_ARGS& parse_args,T kg=1,T s=1,T m=1);
    virtual ~CAHN_HILLIARD();

    virtual void Initialize_Symbolics(std::string code)=0;
    void Initialize_Gamma(const GRID<TV>& grid,ARRAYS_ND_BASE<T,IV>& gamma,int ghost,T dt,T time) const;
    
    virtual void Cahn_Hilliard(T dt,T time,ARRAYS_ND_BASE<T,IV>& gamma,ARRAYS_ND_BASE<T,IV>* gamma_prev=nullptr,ARRAY<T,FACE_INDEX<TV::m> >* face_velocities_ab=nullptr)=0;

    virtual void Register_For_Viewing(VIEWER_OUTPUT &vo)=0;
    
    void Set_Periodic_Boundaries(const VECTOR<bool,TV::m>& p);
    void Set_Neumann_Boundaries(const ARRAY<bool,FACE_INDEX<TV::m> >& N, const ARRAY<T,FACE_INDEX<TV::m> >* N_values=0);
    void Set_Dirichlet_Boundaries(const ARRAYS_ND_BASE<bool,IV>& D, const ARRAYS_ND_BASE<T,IV>* D_values=0);
    void Set_Analytic_Output(bool print_output,bool draw_error,VIEWER_OUTPUT* vo,DEBUG_PARTICLES<TV>* p,HASHTABLE<std::string,DEBUG_PARTICLES<TV>*>* error_p);
   
//#####################################################################
};
}
#endif
