//#####################################################################
// Copyright 2022, Jason Goulding, Mehrnaz Ayazi
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CAHN_HILLIARD_OPTIMIZER__
#define __CAHN_HILLIARD_OPTIMIZER__
#include <Core/Advection/ADVECTION_CONSERVATIVE_ENO.h>
#include <Core/Advection/ADVECTION_CONSERVATIVE_WENO.h>
#include <Core/Arrays/ARRAYS_ND.h>
#include <Core/Arrays/ARRAYS_ND_BASE.h>
#include <Core/Boundaries/BOUNDARY.h>
#include <Core/Boundaries/BOUNDARY_PERIODIC.h>
#include <Core/Data_Structures/HASHTABLE.h>
#include <Core/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Poisson/PROJECTION_UNIFORM.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Core/Vectors/VECTOR.h>
#include <Core/Viewer/VIEWER_OUTPUT.h>
#include "CAHN_HILLIARD.h"
#include <functional>

namespace PhysBAM{

template<class TV> class CHO_SYMBOLIC;

template<class TV>
class CAHN_HILLIARD_OPTIMIZER:public CAHN_HILLIARD<TV>
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<int,TV::m> IV;
    typedef CAHN_HILLIARD<TV> BASE;
    // using BASE::mass_scale; using BASE::time_scale; using BASE::distance_scale;
    using BASE::ghost;
    using BASE::order;
    using BASE::print_gamma_stats;
    using BASE::print_status;
    using BASE::print_solve_results;
    using BASE::print_solve_residual;
    using BASE::debug_output;
    // using BASE::suppress_output; using BASE::error_output;
    // using BASE::viewer_output; using BASE::debug_particles; using BASE::error_particles;
    using BASE::psi_N; using BASE::psi_N_values;
    using BASE::psi_D; using BASE::psi_D_values;
    using BASE::grid;
    using BASE::symbolic;
    // using BASE::periodic_boundary_scalar; using BASE::advection_scalar;
    using BASE::periodic;
    using BASE::side_contact_angle;

    using BASE::sigma;
    using BASE::test_flux;
    using BASE::flux;
    using BASE::linearize;
    // using BASE::test_units; using BASE::discrete_advection;
    using BASE::Gradient;
    using BASE::Laplacian;
    using BASE::Fill_Ghost;
    // using BASE::Discrete_Advection;
    using BASE::Advect;
    using BASE::Print_Gamma_Stats;
    using BASE::Test_Analytic;
    using BASE::Test_Flux;
    // using BASE::Set_Periodic_Boundaries;
    // using BASE::Set_Neumann_Boundaries;
    // using BASE::Set_Dirichlet_Boundaries;
    // using BASE::Set_Analytic_Output;
    using BASE::relative_error;

    ARRAY<T,IV> A;
    ARRAY<T,IV> B;
    ARRAY<T,IV> C;
    ARRAY<T,IV> gamma_star;
    ARRAY<T,IV> gamma_bd;
    ARRAY<T,IV> gamma_ab;
    ARRAY<T,IV>* C_prev=nullptr;
    ARRAY<T,IV>* C_curr=nullptr;
    
    void Solve_Helmholtz_CH_Equation(ARRAYS_ND_BASE<T,IV>& u,const ARRAYS_ND_BASE<T,IV>& f,T beta,const ARRAYS_ND_BASE<T,IV>& gamma_ab,std::function<TV(TV,int,T)> n_func=0) const;
    void Solve_Newton_CH_Equation(ARRAYS_ND_BASE<T,IV>& u,ARRAYS_ND_BASE<T,IV>& v,const ARRAYS_ND_BASE<T,IV>& f,T beta,T k,T dt,const ARRAYS_ND_BASE<T,IV>& gamma_ab,std::function<TV(TV,int,T)> n_func=0) const;
    bool force_q=false;
    T target_q=0.5;
    bool output_intermediates=true;
    bool first_solve=true;

    CHO_SYMBOLIC<TV>* cho_symbolic=nullptr;
    
public:
    using BASE::M;
    using BASE::lambda;
    using BASE::eta;
    bool write_ch_system=false;
    bool use_preconditioner=true;
    bool newton_warm_start=true;
    bool newton_B_start=false;

    T r=0.01;
    T b=0.01;
    T F_scale=0.25;
    T ch_abs_tolerance=1e-30;
    T ch_rel_tolerance=1e-12;
    T ch_n_max_iterations=200;
    T ch_n_abs_tolerance=1e-30;
    T ch_n_rel_tolerance=1e-12;
    T ch_n_k_tolerance=1e-12;
    T ch_n_obj_error=-1;
    T ch_n_ada_tolerance=1e-3;

    bool arbitrary_constant=true;
    bool require_convergence=true;
    bool compute_flux=true;
    bool diff_test=false;
    bool alt_F=true;
    
    CAHN_HILLIARD_OPTIMIZER(const GRID<TV>& g,int ghost,PARSE_ARGS& parse_args,T kg=1,T s=1,T m=1);
    ~CAHN_HILLIARD_OPTIMIZER();

    void Register_For_Viewing(VIEWER_OUTPUT &vo);
    
    void Initialize_Symbolics(std::string code);
    void Cahn_Hilliard(T dt,T time,ARRAYS_ND_BASE<T,IV>& gamma,ARRAYS_ND_BASE<T,IV>* gamma_prev=nullptr,ARRAY<T,FACE_INDEX<TV::m> >* face_velocities_ab=nullptr);   
//#####################################################################
};
}
#endif
