//#####################################################################
// Copyright 2022
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Log/LOG.h>
#include <Core/Matrices/MATRIX.h>
#include <Core/Ordinary_Differential_Equations/RUNGEKUTTA.h>
#include <Core/Symbolics/EXECUTABLE_PROGRAM.h>
#include <Core/Symbolics/PROGRAM_COMPILER.h>
#include <Core/Symbolics/PROGRAM_SCALAR.h>
#include "CHHNS_SYMBOLIC.h"

namespace PhysBAM{

//#####################################################################
// Constructor
//#####################################################################
template<class TV> CHHNS_SYMBOLIC<TV>::
CHHNS_SYMBOLIC(const std::string& analytic_solution,bool linearize,bool use_st,int st_type,bool use_adv,bool fix_momen)
{
    typedef PROGRAM_SCALAR<T> PS;
    PROGRAM_COMPILER<T> pc;
    // pc.debug=true; // 2D variable names
    for(int i=0;i<TV::m;i++)
        idx_X(i)=pc.Add_Input(LOG::sprintf("%c","xyz"[i]).c_str()); //i0-i1?
    idx_t=pc.Add_Input("t"); // i2
    idx_dt=pc.Add_Input("dt"); // i3
    idx_lambda=pc.Add_Input("lambda"); // i4
    idx_eta=pc.Add_Input("eta"); // i5
    idx_M=pc.Add_Input("M"); // i6
    idx_rho0=pc.Add_Input("rho0"); // i7
    idx_rho1=pc.Add_Input("rho1"); // i8
    idx_mu0=pc.Add_Input("mu0"); // i9
    idx_mu1=pc.Add_Input("mu1"); // i10
    for(int i=0;i<TV::m;i++)
        idx_g(i)=pc.Add_Input(LOG::sprintf("g_%c","xyz"[i]).c_str());
    idx_second_order=pc.Add_Input("second_order"); // i13
    idx_t2=pc.Add_Input("t2"); // i14
    idx_t0=pc.Add_Input("t0"); // i15
    idx_sigma=pc.Add_Input("sigma"); // i17
    idx_m=pc.Add_Input("m"); // i18
    idx_s=pc.Add_Input("s"); // i19
    idx_kg=pc.Add_Input("kg"); // i20
    idx_pressure_scale=pc.Add_Input("pressure_scale"); // i21
    idx_speed_scale=pc.Add_Input("speed_scale"); // i22
    input_size=pc.Num_In();
    
    for(int i=0;i<TV::m;i++) pc.Add_Output(LOG::sprintf("%c","uvw"[i]).c_str());
    pc.Add_Output("p");
    pc.Add_Output("gamma");
    auto num_out=pc.Num_Out();
   
    pc.Parse(analytic_solution.c_str());
    auto pc2(pc);
    auto pc0(pc);
    auto pc05(pc);
    pc2.Rename_Variable(idx_t,idx_t2); // time t+dt
    pc0.Rename_Variable(idx_t,idx_t0); // time t-dt
    pc.Append(pc2);
    pc.Append(pc0);
    pc.Append(pc05);
    pc.Optimize();
    
    PS ps_pressure_scale={&pc,idx_pressure_scale};
    PS ps_speed_scale={&pc,idx_speed_scale};
    PS ps_m={&pc,idx_m};
    PS ps_s={&pc,idx_s};
    PS ps_kg={&pc,idx_kg};

    VECTOR<PS,TV::m> ps_u;
    VECTOR<PS,TV::m> ps_u2;
    VECTOR<PS,TV::m> ps_u0;
    VECTOR<PS,TV::m> ps_u05;
    VECTOR<PS,TV::m> ps_zeros;
    for(int i=0;i<TV::m;i++)
    {
        ps_u(i)=Make_Output(PS{&pc,MEM_OUT(i)}*ps_speed_scale);
        ps_u2(i)=Make_Output(PS{&pc,MEM_OUT(i+Value(num_out))}*ps_speed_scale);
        ps_u0(i)=Make_Output(PS{&pc,MEM_OUT(i+2*Value(num_out))}*ps_speed_scale);
        ps_u05(i)=Make_Output(PS{&pc,MEM_OUT(i+3*Value(num_out))}*ps_speed_scale);
        ps_zeros(i)=Make_Output(PS::Number(pc,(T)0));
    }
    Add("u",ps_u);
    Add("u2",ps_u2);
    Add("u0",ps_u0);
    Add("u05",ps_u05);
    Add("zeros",ps_zeros);

    PS ps_p=Make_Output(PS{&pc,MEM_OUT(TV::m)}*ps_pressure_scale);
    PS ps_p2=Make_Output(PS{&pc,MEM_OUT(TV::m+Value(num_out))}*ps_pressure_scale);
    PS ps_p0=Make_Output(PS{&pc,MEM_OUT(TV::m+2*Value(num_out))}*ps_pressure_scale);
    PS ps_p05=Make_Output(PS{&pc,MEM_OUT(TV::m+3*Value(num_out))}*ps_pressure_scale);
    PS ps_gamma={&pc,MEM_OUT(TV::m+1)};
    PS ps_gamma2={&pc,MEM_OUT(TV::m+1+Value(num_out))};
    PS ps_gamma0={&pc,MEM_OUT(TV::m+1+2*Value(num_out))};

    Add("p",ps_p);
    Add("p2",ps_p2);
    Add("p0",ps_p0);
    Add("p05",ps_p05);
    Add("gamma",ps_gamma);
    Add("gamma2",ps_gamma2);
    Add("gamma0",ps_gamma0);
    
    PS ps_dt={&pc,idx_dt};
    PS ps_lambda={&pc,idx_lambda};
    PS ps_eta={&pc,idx_eta};
    PS ps_M={&pc,idx_M};
    PS ps_rho0={&pc,idx_rho0};
    PS ps_rho1={&pc,idx_rho1};
    PS ps_mu0={&pc,idx_mu0};
    PS ps_mu1={&pc,idx_mu1};
    VECTOR<PS,TV::m> ps_g;
    for(int i=0;i<TV::m;i++) ps_g(i)={&pc,idx_g(i)};
    Add("g",ps_g);
    PS ps_second_order={&pc,idx_second_order};
    PS ps_t2={&pc,idx_t2};
    PS ps_t0={&pc,idx_t0};
    PS ps_sigma={&pc,idx_sigma};

    (void)ps_dt;
    (void)ps_lambda;
    (void)ps_eta;
    (void)ps_M;
    (void)ps_rho0;
    (void)ps_rho1;
    (void)ps_mu0;
    (void)ps_mu1;
    (void)ps_second_order;
    (void)ps_t2;
    (void)ps_t0;
    (void)ps_sigma;
    (void)ps_m;
    (void)ps_s;
    (void)ps_kg;
    
    VECTOR<MEM_IN,TV::m> axii;
    for(int i=0;i<TV::m;i++) axii(i)=MEM_IN(i);

#define MAKE(name,ps) auto ps_##name=Make_Output(ps);Add(#name,ps_##name);
    MAKE(2,PS::Number(pc,(T)2));
    MAKE(1,PS::Number(pc,(T)1));
    MAKE(0,PS::Number(pc,(T)0));
    MAKE(epsilon,PS::Number(pc,(T)1e-8));
    
    auto dF=[linearize,&ps_2,&ps_eta,&ps_1](const auto& g)
    {
        if(linearize) return g/(ps_eta*ps_eta);
        else return (ps_2*(ps_2*g*g*g-(ps_2+ps_1)*g*g+g))/(ps_eta*ps_eta);
    };
    MAKE(order,ps_1+ps_second_order/ps_2);
    MAKE(dt_hat,ps_dt/ps_order);
    MAKE(adv,use_adv?ps_1:ps_0);
    MAKE(gamma_ab,ps_second_order*(ps_gamma-ps_gamma0)+ps_gamma);
    MAKE(gamma_bd,(ps_second_order*(ps_gamma-ps_gamma0/ps_2)+ps_gamma)/ps_order);
    MAKE(density,ps_rho0*(ps_1-ps_gamma)+ps_rho1*ps_gamma);
    MAKE(viscosity,ps_mu0*(ps_1-ps_gamma)+ps_mu1*ps_gamma);
    MAKE(density2,ps_rho0*(ps_1-ps_gamma2)+ps_rho1*ps_gamma2);
    MAKE(viscosity2,ps_mu0*(ps_1-ps_gamma2)+ps_mu1*ps_gamma2);
    MAKE(density0,ps_rho0*(ps_1-ps_gamma0)+ps_rho1*ps_gamma0);
    MAKE(rho_u_ab,ps_second_order*(ps_density*ps_u-ps_density0*ps_u0)+ps_density*ps_u);
    MAKE(rho_u_bd,(ps_second_order*(ps_density*ps_u-ps_density0*ps_u0/ps_2)+ps_density*ps_u)/ps_order);
    MAKE(rho_ab,ps_second_order*(ps_density-ps_density0)+ps_density);
    MAKE(rho_bd,(ps_second_order*(ps_density-ps_density0/ps_2)+ps_density)/ps_order);
    MAKE(u_ab,ps_second_order*(ps_u-ps_u0)+ps_u);
    MAKE(u_bd,(ps_second_order*(ps_u-ps_u0/ps_2)+ps_u)/ps_order);
    MAKE(S,sqrt(ps_M*ps_lambda/ps_dt_hat));
    MAKE(a,ps_M*ps_lambda/ps_S);
    MAKE(b,ps_2);
    MAKE(grad_gamma,Gradient(ps_gamma,axii));
    MAKE(grad_gamma_ab,Gradient(ps_gamma_ab,axii));
    MAKE(grad_gamma_bd,Gradient(ps_gamma_bd,axii));
    MAKE(grad_gamma2,Gradient(ps_gamma2,axii));
    MAKE(grad_gamma0,Gradient(ps_gamma0,axii));
    MAKE(gamma_star,ps_gamma_bd-ps_dt_hat*ps_adv*Divergence(ps_u_ab*ps_gamma_ab,axii));
    // MAKE(density_star,ps_rho0+(ps_rho1-ps_rho0)*ps_gamma_star);
    MAKE(grad_gamma_star,Gradient(ps_gamma_star,axii));
    MAKE(gamma_D,ps_gamma2-ps_S*ps_dt_hat*Laplacian(ps_gamma2,axii));
    MAKE(gamma_t,Diff(ps_gamma_ab,idx_t));
    MAKE(gamma_A,ps_a*dF(ps_gamma_ab)-ps_b*ps_gamma_ab);
    MAKE(gamma_C,ps_gamma_D+ps_gamma_A);
    MAKE(grad_gamma_C,Gradient(ps_gamma_C,axii));
    MAKE(xi,ps_lambda*(dF(ps_gamma_ab)-Laplacian(ps_gamma2,axii))+ps_b*ps_S/ps_M*(ps_gamma2-ps_gamma_ab));
    // MAKE(xi,ps_lambda*(dF(ps_gamma_ab)-Laplacian(ps_gamma2,axii)));
    MAKE(grad_xi,Gradient(ps_xi,axii));
    MAKE(src_difft,Diff(ps_gamma2,idx_t2));
    MAKE(src_inert,Diff(ps_gamma2,idx_t2)/ps_s+ps_adv*Divergence(ps_u2*ps_gamma2,axii)/ps_m);
    MAKE(src_xi,ps_lambda*(dF(ps_gamma2)-Laplacian(ps_gamma2,axii)/(ps_m*ps_m)));
    MAKE(src_ch,ps_src_inert-ps_M*Laplacian(ps_src_xi,axii)/(ps_m*ps_m));
    
    MAKE(gamma_B,ps_gamma_star+ps_gamma_A+ps_dt_hat*ps_src_ch);
    // MAKE(src_mass_flux,ps_density2*ps_u2-(ps_rho1-ps_rho0)*(ps_M*Gradient(ps_src_xi,axii)+ps_src_flux_ch));
    MAKE(src_mass_flux,ps_density2*ps_u2-(ps_rho1-ps_rho0)*ps_M*Gradient(ps_src_xi,axii));
    MAKE(src_ns_inert_part1,Diff(ps_density2*ps_u2,idx_t2)/ps_s);
    MAKE(src_ns_inert_part2,ps_adv*Divergence_L(Outer_Product(ps_src_mass_flux,ps_u2),axii)/ps_m);
    MAKE(src_ns_inert,ps_src_ns_inert_part1+ps_src_ns_inert_part2);
    MAKE(src_ns_p,Gradient(ps_p2,axii)/ps_m);
    MAKE(st,use_st?ps_1:ps_0);
    MAKE(st0,ps_st*ps_lambda*Divergence_L(Outer_Product(ps_grad_gamma2,ps_grad_gamma2),axii)/ps_m);
    MAKE(st1,ps_st*ps_2*ps_2*ps_src_xi*Gradient(ps_gamma2,axii)/ps_m);
    MAKE(g_delta,ps_gamma2);
    MAKE(mag_grad_gamma2,ps_grad_gamma2.Magnitude());
    MAKE(src_curv,cond_lazy(ps_mag_grad_gamma2>ps_0,[&](){return -Divergence(ps_grad_gamma2/ps_grad_gamma2.Magnitude(),axii);},[&](){return ps_0;}));
    MAKE(st2,ps_st*ps_sigma*ps_src_curv*Gradient(ps_g_delta,axii));
    MAKE(src_ns_st,st_type==0?ps_st0:(st_type==1?ps_st1:ps_st2));
    MAKE(adv_flux,ps_rho_ab*ps_u_ab);
    MAKE(ch_flux,(ps_rho1-ps_rho0)*ps_M*ps_grad_xi);
    // MAKE(ch_flux,(ps_rho1-ps_rho0)*(ps_S*Gradient(ps_gamma_C,axii)+ps_S*Gradient(ps_gamma2,axii)+ps_src_flux_ch));
    MAKE(mass_flux,ps_adv_flux-ps_ch_flux);
    // MAKE(mass_flux,ps_rho_u_ab-(ps_rho1-ps_rho0)*(ps_M*ps_grad_xi));
    MAKE(rho_u_star,ps_rho_u_bd-ps_dt_hat*ps_adv*Divergence_L(Outer_Product(ps_mass_flux,ps_u_ab),axii)); // Add advection
    MAKE(u_star,ps_rho_u_star/ps_density2); // Momentum fix via NS equation
    MAKE(src_ns_visc,Divergence_L((ps_viscosity2/ps_m)*(Gradient(ps_u2,axii)+Gradient(ps_u2,axii).Transposed()),axii)/ps_m);
    MAKE(src_ns,ps_src_ns_inert+ps_src_ns_p-ps_density2*ps_g-ps_src_ns_st-ps_src_ns_visc);
    
    MAKE(grad_p,Gradient(ps_p,axii));
    MAKE(grad_p0,Gradient(ps_p0,axii));
    // MAKE(grad_u,Gradient(ps_u,axii));
    MAKE(grad_u,Gradient(ps_u(0),axii));
    MAKE(grad_v,Gradient(ps_u(1),axii));
    if(TV::m>2) {MAKE(grad_w,Gradient(ps_u(2),axii));}
    MAKE(grad_u0,Gradient(ps_u0(0),axii));
    MAKE(grad_v0,Gradient(ps_u0(1),axii));
    if(TV::m>2) {MAKE(grad_w0,Gradient(ps_u0(2),axii));}
    MAKE(grad_u_star,Gradient(ps_u_star(0),axii));
    MAKE(grad_v_star,Gradient(ps_u_star(1),axii));
    if(TV::m>2) {MAKE(grad_w_star,Gradient(ps_u_star(2),axii));}
    MAKE(grad_p2,Gradient(ps_p2,axii));
    // MAKE(grad_u2,Gradient(ps_u2,axii));
    MAKE(grad_u2,Gradient(ps_u2(0),axii));
    MAKE(grad_v2,Gradient(ps_u2(1),axii));
    if(TV::m>2) {MAKE(grad_w2,Gradient(ps_u2(2),axii));}
    
    MAKE(u_star_star,ps_u_star+ps_dt_hat*(ps_g+ps_src_ns_st/ps_density2+ps_src_ns/ps_density2)); // Add forces
    MAKE(visc,Divergence_L(ps_viscosity2*Gradient(ps_u_ab,axii),axii));
    MAKE(u_star_star_visc,ps_u_star_star+ps_dt_hat*ps_visc/ps_density2); // Add visc_div_grad_t
    MAKE(u_visc_prev,ps_u_star_star_visc-ps_dt_hat*ps_grad_p/ps_density2);
    MAKE(test_gamma,ps_gamma_ab-ps_gamma_ab*ps_dt_hat/(ps_a*ps_dt_hat+ps_2*ps_dt_hat*ps_S+ps_1));
    MAKE(p_avg,(ps_p2+ps_p)/ps_2);
    MAKE(p_diff,ps_p2-ps_p);
    MAKE(grad_p_prime,Gradient(ps_p_diff,axii));
    MAKE(u_visc,ps_u2+ps_dt_hat*ps_grad_p_prime/ps_density2);
    MAKE(grad_u_visc,Gradient(ps_u_visc(0),axii));
    MAKE(grad_v_visc,Gradient(ps_u_visc(1),axii));
    if(TV::m>2) {MAKE(grad_w_visc,Gradient(ps_u_visc(2),axii));}

    MAKE(u_diff,ps_u2-ps_u);

    pc.Optimize();
#undef MAKE
}
//#####################################################################
// Destructor
//#####################################################################
template<class TV> CHHNS_SYMBOLIC<TV>::
~CHHNS_SYMBOLIC()
{
}
//#####################################################################
// Function Add
//#####################################################################
template<class TV> void CHHNS_SYMBOLIC<TV>::
Add(const std::string& name,const PROGRAM_SCALAR<T>& ps)
{
    t_map.Set(name,new EXECUTABLE_PROGRAM<T>(ps));
}
//#####################################################################
// Function Add
//#####################################################################
template<class TV> template<int d> void CHHNS_SYMBOLIC<TV>::
Add(const std::string& name,const VECTOR<PROGRAM_SCALAR<T>,d>& ps)
{
    tv_map.Set(name,new EXECUTABLE_PROGRAM<TV>(ps));
}
//#####################################################################
// Function Add
//#####################################################################
// template<class TV> template<int m,int n> void CHHNS_SYMBOLIC<TV>::
// Add(const std::string& name,const MATRIX<PROGRAM_SCALAR<T>,m,n>& ps)
// {
//     mat_map.Set(name,new EXECUTABLE_PROGRAM<MATRIX<T,TV::m,TV::m> >(ps));
// }
//#####################################################################
// Function Set_Step
//#####################################################################
template<class TV> void CHHNS_SYMBOLIC<TV>::
Set_Step(T time, T dt)
{
    inputs.Resize(input_size,init_all);
    inputs.Subset(idx_X)=TV();
    inputs(idx_t)=time/time_scale;
    inputs(idx_dt)=dt/time_scale;
    inputs(idx_lambda)=lambda;
    inputs(idx_eta)=eta;
    inputs(idx_M)=M;
    inputs(idx_rho0)=rho0;
    inputs(idx_rho1)=rho1;
    inputs(idx_mu0)=mu0;
    inputs(idx_mu1)=mu1;
    inputs.Subset(idx_g)=gravity;
    inputs(idx_second_order)=second_order;
    inputs(idx_t2)=(time+dt)/time_scale;
    inputs(idx_t0)=(time-dt)/time_scale;
    inputs(idx_sigma)=sigma;
    inputs(idx_m)=distance_scale;
    inputs(idx_s)=time_scale;
    inputs(idx_kg)=mass_scale;
    inputs(idx_pressure_scale)=pressure_scale;
    inputs(idx_speed_scale)=speed_scale;
}
template class CHHNS_SYMBOLIC<VECTOR<float,2> >;
template class CHHNS_SYMBOLIC<VECTOR<double,2> >;
template class CHHNS_SYMBOLIC<VECTOR<float,3> >;
template class CHHNS_SYMBOLIC<VECTOR<double,3> >;
}
