//#####################################################################
// Copyright 2024.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Diff_Tools/Diff_Test.h>
#include <Core/Krylov_Solvers/KRYLOV_VECTOR_WRAPPER.h>
#include <Core/Log/LOG.h>
#include <Core/Math_Tools/pow.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include "CHN_OBJECTIVE.h"
namespace PhysBAM{
//#####################################################################
// Helpers
//####################################################################
template<typename T> T dh0(T x,T r,T k,bool alt_F)
{
    if(alt_F)
    {
        T p=1+6*k*(1-4*r);
        T L=48*k*pow<3>(r)*p;
        T s=sqrt(x*x+L);
        if(x>=0) return x+s;
        return -L/(x-s);
    }
    else
    {
        T q=6*k*(1-2*r),qrr=q*r*r,s=sqrt(x*x+qrr);
        if(x>=0) return -qrr/(x+s);
        return x-s;
    }
}
template<typename T> T h0(T x,T r,T k,bool alt_F)
{
    if(alt_F)
    {
        T p=1+6*k*(1-4*r);
        T L=48*k*pow<3>(r)*p;
        T z=dh0(x,r,k,alt_F);
        return z*x/2+(L/2)*log(z);
    }
    else
    {
        T q=6*k*(1-2*r);
        T z=dh0(x,r,k,alt_F);
        return x*z/2+(q*r*r)*log(-z)/2; // test dh0(-x) instead of dh0(x) --- Appears correct
    }
}
template<typename T> T g0(T x,T r,T k,bool alt_F)
{
    if(alt_F)
    {
        T s=6*r*(1-6*r)*k;
        T p=1+6*k*(1-4*r);
        return h0(x+s,r,k,alt_F)/(2*p);
    }
    else
    {
        T q=6*k*(1-2*r);
        return 2*h0((q*r-x)/2,r,k,alt_F);
    }
}
template<typename T> T h1(T x,T r,T k)
{
    T q=6*k*(1-2*r);
    T K=(1-q)/(24*k);
    return K*x+(T)2/3*pow<3,2>(K*K-x);
}
template<typename T> T g1(T x,T r,T k)
{
    return x/2+12*k*h1((x-(T)0.5)/(12*k),r,k);
}
template<typename T> T dg0(T x,T r,T k,bool alt_F)
{
    if(alt_F)
    {
        T s=6*r*(1-6*r)*k;
        T p=1+6*k*(1-4*r);
        return dh0(x+s,r,k,alt_F)/(2*p);
    }
    else
    {
        T q=6*k*(1-2*r);
        return -dh0((q*r-x)/2,r,k,alt_F);
    }
}
template<typename T> T dh1(T x,T r,T k)
{
    T q=6*k*(1-2*r);
    T K=(1-q)/(24*k);
    return x/(K+sqrt(K*K-x));
}
template<typename T> T dg1(T x,T r,T k)
{
    return (T)0.5+dh1((x-(T)0.5)/(12*k),r,k);
}
template<typename T> T ddh0(T x,T r,T k,bool alt_F)
{
    if(alt_F)
    {
        T p=1+6*k*(1-4*r);
        T L=48*k*pow<3>(r)*p;
        return 1+x/sqrt(x*x+L);
    }
    else
    {
        T q=6*k*(1-2*r);
        return 1-x/sqrt(x*x+q*r*r);
    }
}
template<typename T> T ddg0(T x,T r,T k,bool alt_F)
{
    if(alt_F)
    {
        T s=6*r*(1-6*r)*k;
        T p=1+6*k*(1-4*r);
        return ddh0(x+s,r,k,alt_F)/(2*p);
    }
    else
    {
        T q=6*k*(1-2*r);
        return ddh0((q*r-x)/2,r,k,alt_F)/2;
    }
}
template<typename T> T ddh1(T x,T r,T k)
{
    T q=6*k*(1-2*r);
    T K=(1-q)/(24*k);
    return 1/(2*sqrt(K*K-x));  // test negation
}
template<typename T> T ddg1(T x,T r,T k)
{
    return ddh1((x-(T)0.5)/(12*k),r,k)/(12*k);
}


//#####################################################################
// Constructor
//#####################################################################
template<class TV> CHN_OBJECTIVE<TV>::
CHN_OBJECTIVE(SPARSE_MATRIX_FLAT_MXN<T>& A,const ARRAY<T> &b_in,T k_in,T r_in,bool enforce,bool altF,bool linear)
    :enforce_compatibility(enforce),r(r_in),k(k_in),alt_F(altF),linearized(linear),L(A),system(mod_L),b(b_in)
{

}
//#####################################################################
// Destructor
//#####################################################################
template<class TV> CHN_OBJECTIVE<TV>::
~CHN_OBJECTIVE()
{
    system.nullspace_vectors.Delete_Pointers_And_Clean_Memory();
}
//#####################################################################
// g
//#####################################################################
template<class TV> auto CHN_OBJECTIVE<TV>::
g(T x) const -> T
{
    if(linearized) return sqr(x)/(2+2*k);
    if(x<=r) // assumes x>=0
        return g0(x,r,k,alt_F)-g0(r,r,k,alt_F)+g1(r,r,k);
    else if(x<(T)0.5)
        return g1(x,r,k);
    else if(x<=(T)(1-r))
        return x-(T)0.5+g1(1-x,r,k);
    else // assumes x<=1
        return x-(T)0.5+g0(1-x,r,k,alt_F)-g0(r,r,k,alt_F)+g1(r,r,k); 
}
//#####################################################################
// g
//#####################################################################
template<class TV> auto CHN_OBJECTIVE<TV>::
dg(T x) const -> T
{
    if(linearized) return x/(1+k);
    if(x<=r) // assumes x>=0
        return dg0(x,r,k,alt_F);
    else if(x<(T)0.5)
        return dg1(x,r,k);
    else if(x<=(T)(1-r))
        return (T)1-dg1(1-x,r,k);
    else // assumes x<=1
        return (T)1-dg0(1-x,r,k,alt_F);
}
//#####################################################################
// g
//#####################################################################
template<class TV> auto CHN_OBJECTIVE<TV>::
ddg(T x) const -> T
{
    if(linearized) return 1/(1+k);
    if(x<=r) // assumes x>=0
        return ddg0(x,r,k,alt_F);
    else if(x<(T)0.5)
        return ddg1(x,r,k);
    else if(x<=(T)(1-r))
        return ddg1(1-x,r,k);
    else // assumes x<=1
        return ddg0(1-x,r,k,alt_F);
}
//#####################################################################
// Destructor
//#####################################################################
template<class TV> void CHN_OBJECTIVE<TV>::
Test_g() const
{
    T q=6*k*(1-2*r);
    T K=(1-q)/(24*k);
    LOG::printf("q=%P K=%P\n",q,K);
    RANDOM_NUMBERS<T> rand(1234);
    T eps=1e-6;
    T x0=rand.Get_Uniform_Number(0+eps,r-eps); // random in range 0 to r
    T x1=x0+eps;
    Diff_Test(0,"g_a",g(x0),x0,dg(x0));
    Diff_Test(1,"g_a",g(x1),x1,dg(x1));
    Diff_Test(0,"g0_a",g0(x0,r,k,alt_F),x0,dg0(x0,r,k,alt_F));
    Diff_Test(1,"g0_a",g0(x1,r,k,alt_F),x1,dg0(x1,r,k,alt_F));
    Diff_Test(0,"h0_a",h0(x0,r,k,alt_F),x0,dh0(x0,r,k,alt_F));
    Diff_Test(1,"h0_a",h0(x1,r,k,alt_F),x1,dh0(x1,r,k,alt_F));
    Diff_Test(0,"dg_a",dg(x0),x0,ddg(x0));
    Diff_Test(1,"dg_a",dg(x1),x1,ddg(x1));
    Diff_Test(0,"dg0_a",dg0(x0,r,k,alt_F),x0,ddg0(x0,r,k,alt_F));
    Diff_Test(1,"dg0_a",dg0(x1,r,k,alt_F),x1,ddg0(x1,r,k,alt_F));
    Diff_Test(0,"dh0_a",dh0(x0,r,k,alt_F),x0,ddh0(x0,r,k,alt_F));
    Diff_Test(1,"dh0_a",dh0(x1,r,k,alt_F),x1,ddh0(x1,r,k,alt_F));

    x0=rand.Get_Uniform_Number(r+eps,0.5-eps);; // random in range r to 0.5
    x1=x0+eps;
    Diff_Test(0,"g_b",g(x0),x0,dg(x0));
    Diff_Test(1,"g_b",g(x1),x1,dg(x1));
    Diff_Test(0,"g1_b",g1(x0,r,k),x0,dg1(x0,r,k));
    Diff_Test(1,"g1_b",g1(x1,r,k),x1,dg1(x1,r,k));
    Diff_Test(0,"h1_b",h1(x0,r,k),x0,dh1(x0,r,k));
    Diff_Test(1,"h1_b",h1(x1,r,k),x1,dh1(x1,r,k));
    Diff_Test(0,"dg_b",dg(x0),x0,ddg(x0));
    Diff_Test(1,"dg_b",dg(x1),x1,ddg(x1));
    Diff_Test(0,"dg1_b",dg1(x0,r,k),x0,ddg1(x0,r,k));
    Diff_Test(1,"dg1_b",dg1(x1,r,k),x1,ddg1(x1,r,k));
    Diff_Test(0,"dh1_b",dh1(x0,r,k),x0,ddh1(x0,r,k));
    Diff_Test(1,"dh1_b",dh1(x1,r,k),x1,ddh1(x1,r,k));

    x0=rand.Get_Uniform_Number(0.5+eps,1-r-eps);; // random in range 0.5 to 1-r
    x1=x0+eps;
    Diff_Test(0,"g_c",g(x0),x0,dg(x0));
    Diff_Test(1,"g_c",g(x1),x1,dg(x1));
    Diff_Test(0,"g1_c",g1(x0,r,k),x0,dg1(x0,r,k));
    Diff_Test(1,"g1_c",g1(x1,r,k),x1,dg1(x1,r,k));
    Diff_Test(0,"h1_c",h1(x0,r,k),x0,dh1(x0,r,k));
    Diff_Test(1,"h1_c",h1(x1,r,k),x1,dh1(x1,r,k));
    Diff_Test(0,"dg_c",dg(x0),x0,ddg(x0));
    Diff_Test(1,"dg_c",dg(x1),x1,ddg(x1));
    Diff_Test(0,"dg1_c",dg1(x0,r,k),x0,ddg1(x0,r,k));
    Diff_Test(1,"dg1_c",dg1(x1,r,k),x1,ddg1(x1,r,k));
    Diff_Test(0,"dh1_c",dh1(x0,r,k),x0,ddh1(x0,r,k));
    Diff_Test(1,"dh1_c",dh1(x1,r,k),x1,ddh1(x1,r,k));

    x0=rand.Get_Uniform_Number(1-r+eps,1-eps);; // random in range 1-r to 1
    x1=x0+eps;
    Diff_Test(0,"g_d",g(x0),x0,dg(x0));
    Diff_Test(1,"g_d",g(x1),x1,dg(x1));
    Diff_Test(0,"g0_d",g0(x0,r,k,alt_F),x0,dg0(x0,r,k,alt_F));
    Diff_Test(1,"g0_d",g0(x1,r,k,alt_F),x1,dg0(x1,r,k,alt_F));
    Diff_Test(0,"h0_d",h0(x0,r,k,alt_F),x0,dh0(x0,r,k,alt_F));
    Diff_Test(1,"h0_d",h0(x1,r,k,alt_F),x1,dh0(x1,r,k,alt_F));
    Diff_Test(0,"dg_d",dg(x0),x0,ddg(x0));
    Diff_Test(1,"dg_d",dg(x1),x1,ddg(x1));
    Diff_Test(0,"dg0_d",dg0(x0,r,k,alt_F),x0,ddg0(x0,r,k,alt_F));
    Diff_Test(1,"dg0_d",dg0(x1,r,k,alt_F),x1,ddg0(x1,r,k,alt_F));
    Diff_Test(0,"dh0_d",dh0(x0,r,k,alt_F),x0,ddh0(x0,r,k,alt_F));
    Diff_Test(1,"dh0_d",dh0(x1,r,k,alt_F),x1,ddh0(x1,r,k,alt_F));
}
//#####################################################################
// Function Compute
//#####################################################################
template<class TV> void CHN_OBJECTIVE<TV>::
Compute(const KRYLOV_VECTOR_BASE<T>& x_in,KRYLOV_SYSTEM_BASE<T>* hess,KRYLOV_VECTOR_BASE<T>* grad,T* e) const
{
    const KRYLOV_VECTOR_WRAPPER<T,ARRAY<T>>& x=debug_cast<const KRYLOV_VECTOR_WRAPPER<T,ARRAY<T>>&>(x_in);
    T q=6*k*(1-2*r);
    PHYSBAM_ASSERT(q>0 && q<1);
    ARRAY<T> t(x.v.m);
    if(e) // Evaluate Objective Function
    {
        // e=w^T*g(x)-1/2*x^T*A*x-x^T*B ?
        // e=w^T*g(x)-1/2*(A^T*x)^T*x-x^T*B
        T g_sum=0;
        for(int i=0;i<x.v.m;i++)
            g_sum+=g(x.v(i));

        L.Transpose_Times(x.v,t);
        *e=g_sum-0.5*t.Dot(x.v)-x.v.Dot(b);
        // LOG::printf("OBJ: %P %P %P %P\n",x.v.m,g_sum,0.5*t.Dot(x.v),x.v.Dot(b));
    }
    if(grad) // Evaluate Objective Gradient
    {
        // g=g'(C)-A*x-B ?
        L.Times(x.v,t);
        KRYLOV_VECTOR_WRAPPER<T,ARRAY<T>>* gv=debug_cast<KRYLOV_VECTOR_WRAPPER<T,ARRAY<T>>*>(grad);
        assert(gv->v.m==x.v.m);
        for(int i=0;i<b.m;i++)
            gv->v(i)=dg(x.v(i))-t(i)-b(i);
    }
    if(hess) // Evaluate Objective Hessian
    {
        mod_L=L;
        mod_L*=-1;
        mod_L.Initialize_Diagonal_Index();
        assert(mod_L.diagonal_index.m==x.v.m);
        for(int i=0;i<mod_L.diagonal_index.m;i++)
        {
            int diagonal=mod_L.diagonal_index(i);
            T diag=ddg(x.v(i));
            mod_L.A(diagonal).a+=diag;
            PHYSBAM_ASSERT(diag>=0);
        }

        MATRIX_SYSTEM<SPARSE_MATRIX_FLAT_MXN<T>,T,KRYLOV_VECTOR_WRAPPER<T,ARRAY<T>>>* sys=debug_cast<MATRIX_SYSTEM<SPARSE_MATRIX_FLAT_MXN<T>,T,KRYLOV_VECTOR_WRAPPER<T,ARRAY<T>>>*>(hess);
        // sys->A=mod_L;

        // Set up preconditioner
        temp.v.Resize(x.v.m);
        // Use incomplete_cholesky
        
        if(use_preconditioner==true)
        {
            sys->use_preconditioner=true;
            sys->preconditioner_commutes_with_projection=false;

            // recompute_preconditioner
            T modified_incomplete_cholesky_coefficient=(T).97;
            T preconditioner_zero_tolerance=(T)1e-8;
            T preconditioner_zero_replacement=(T)1e-8;
            bool modified_incomplete_cholesky=true;
            mod_L.Construct_Incomplete_Cholesky_Factorization(modified_incomplete_cholesky,
                    modified_incomplete_cholesky_coefficient,preconditioner_zero_tolerance,
                    preconditioner_zero_replacement);
            sys->Set_Preconditioner(*(sys->A.C),temp);
        }

        // MATRIX_SYSTEM<SPARSE_MATRIX_FLAT_MXN<T>,T,KRYLOV_VECTOR_WRAPPER<T,ARRAY<T> > > sys(mod_L);
        // *system=sys;
    }
}
template class CHN_OBJECTIVE<VECTOR<double,2> >;
template class CHN_OBJECTIVE<VECTOR<double,3> >;
template class CHN_OBJECTIVE<VECTOR<float,2> >;
template class CHN_OBJECTIVE<VECTOR<float,3> >;
}
