//##############################################################################
// MARCH2023-Mehrnaz Ayazi
//##############################################################################
#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays/FACE_ARRAYS.h>
#include <Core/Grids/GRID.h>
#include <Core/Grids_Uniform_Computations/MARCHING_CUBES.h>
#include <Core/Images/TEX_FILE.h>
#include <Core/Interpolation/AVERAGING_UNIFORM.h>
#include <Core/Math_Tools/RANGE.h>
#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Core/Meshes/MESH_OBJECT.h>
#include <Core/Meshes/SEGMENTED_CURVE.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Utilities/PROCESS_UTILITIES.h>
#include <Core/Utilities/VIEWER_DIR.h>

using namespace PhysBAM;

typedef double T;
typedef VECTOR<T,2> TV;
typedef VECTOR<int,2> IV;

std::string fig_dir;
ARRAY<std::string> dir;
ARRAY<T> dt;
int n=5;
//#####################################################################
// Function Dump_streamline_data
//#####################################################################
void Dump_streamline_data(const std::string& file)
{
    for(int i=0;i<dir.m;i++)
    {
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        viewer_dir.Read_Last_Frame(0);

        GRID<TV> grid;
        ARRAY<T,IV> gamma;
        ARRAY<T,FACE_INDEX<TV::m>> v;

        Read_From_File(viewer_dir.Common_File("grid"),grid);
        Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);
        Read_From_File(viewer_dir.Frame_File("velocities"),v);
        TV grid_length=grid.Domain().Edge_Lengths();
        TV grid_left_corner=grid.Domain().min_corner;
        TV step_size=grid_length/n;
        for(int j=0;j<n;j++)
        {
            LOG::cout<<"inside the for loop "<<"x0_"+std::to_string(j)+dir(i)+"_"+file<<std::endl;
            std::ofstream stream(fig_dir+"/txt/x0_"+std::to_string(j)+dir(i)+"_"+file);
            LOG::fprintf(stream,"x y\n");
            T x=0;
            T y=step_size(1)*j;

            while(grid.Valid_Cell(grid.Cell_I({x,y}))) {
                LOG::fprintf(stream,"%p %p\n",x,y);
                auto index=grid.Cell_I({x,y});
                AVERAGING_UNIFORM<TV> au;
                auto cell_v=au.Face_To_Cell_Vector(grid,index,v);
                x+=cell_v(0)*dt(i);
                y+=cell_v(1)*dt(i);
            }
        }
        for(int j=0;j<n;j++)
        {
            LOG::cout<<"inside the for loop "<<"y0_"+std::to_string(j)+dir(i)+"_"+file<<std::endl;
            std::ofstream stream(fig_dir+"/txt/y0_"+std::to_string(j)+dir(i)+"_"+file);
            LOG::fprintf(stream,"x y\n");
            T x=step_size(1)*j;
            T y=0;
            while(grid.Valid_Cell(grid.Cell_I({x,y}))) {
                LOG::fprintf(stream,"%p %p\n",x,y);
                auto index=grid.Cell_I({x,y});
                AVERAGING_UNIFORM<TV> au;
                auto cell_v=au.Face_To_Cell_Vector(grid,index,v);
                x+=cell_v(0)*dt(i);
                y+=cell_v(1)*dt(i);
            }
        }
    }
}
//#####################################################################
// Function Dump_Surface_data
//#####################################################################
void Dump_Surface_data(const std::string& file,bool last_frame=true)
{
    for(int i=0;i<dir.m;i++)
    {

        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        GRID<TV> grid;
        ARRAY<T,IV> gamma;

        Read_From_File(viewer_dir.Common_File("grid"),grid);
        std::string file_num=last_frame?"1":"0";
//        std::ofstream stream(fig_dir+"/txt/advection_surface"+file_num+"_"+dir(i)+"_"+file);
        std::ofstream stream(fig_dir+"/txt/advection_surface"+file_num+"_"+dir(i)+"_"+file);
        if(last_frame)
        {
            viewer_dir.Read_Last_Frame(0);
        }
        else
        {
            viewer_dir.Set(0);
        }
        LOG::fprintf(stream,"x y\n");


        Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);

        //create surface and add to text file
        MESH_OBJECT<TV,TV::m-1> surface;
        T contour_value=(T)0.5;
        MARCHING_CUBES<TV>::Create_Surface(surface,grid,at_cells,gamma,contour_value);
        surface.mesh.Initialize_Ordered_Components();
        bool is_surface=true;
        for(const auto &l:*surface.mesh.ordered_loop_nodes)
        {
            if(is_surface){
                ARRAY<TV> pts(surface.particles.X.Subset(l));
                for (int j=0;j<pts.Size();++j) {
                    LOG::fprintf(stream,"%p %p\n",pts(j)(0),pts(j)(1));
                }
            }
            is_surface=!is_surface;
        }
    }
}
//#####################################################################
// Function main
//#####################################################################
int main(int argc,char *argv[])
{
    dt.Fill(1/24);
    PROCESS_UTILITIES::Set_Floating_Point_Exception_Handling(true);
    PROCESS_UTILITIES::Set_Backtrace(true);

    PARSE_ARGS parse_args(argc,argv);
    LOG::Initialize_Logging(false,false,1<<30,true);
    LOG::cout<<parse_args.Print_Arguments()<<std::endl;

    VECTOR<T,2> range_scale={0,1};
    std::string boundary_file="";

    parse_args.Add("-ob",&boundary_file,"string-file name","file name for data");
    parse_args.Add("-i",&dir,"string-directory name","directory name to read data from");
    parse_args.Add("-dt",&dt,"scalar-array","frame_dt");
    parse_args.Add("-n",&n,"scalar","# of stream lines on each side");
    parse_args.Add("-fig",&fig_dir,"dir","base figure directory");
    parse_args.Parse();

    Dump_streamline_data(boundary_file+".txt");
    Dump_Surface_data(boundary_file+".txt");
    Dump_Surface_data(boundary_file+".txt",false);
    LOG::Finish_Logging();
    return 0;
}
