//##############################################################################
// MARCH2023-Mehrnaz Ayazi
//##############################################################################
#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays/FACE_ARRAYS.h>
#include <Core/Grids/GRID.h>
#include <Core/Grids_Uniform_Computations/MARCHING_CUBES.h>
#include <Core/Images/TEX_FILE.h>
#include <Core/Interpolation/AVERAGING_UNIFORM.h>
#include <Core/Math_Tools/RANGE.h>
#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Core/Meshes/MESH_OBJECT.h>
#include <Core/Meshes/SEGMENTED_CURVE.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Utilities/PROCESS_UTILITIES.h>
#include <Core/Utilities/VIEWER_DIR.h>

using namespace PhysBAM;

typedef double T;
typedef VECTOR<T,2> TV;
typedef VECTOR<int,2> IV;

std::string fig_dir;
T dt=0.008;
T r=0;
bool table=false;
bool ref=false;
//#####################################################################
// Function Dump_Surface_data
//#####################################################################
void Dump_Surface_data(const ARRAY<std::string> &dir)
{
    for(int i=0;i<dir.m;i++)
    {
        LOG::cout<<"creating data file "<<fig_dir+"/txt/surface_"+dir(i)+".txt"<<std::endl;
        std::ofstream stream(fig_dir+"/txt/surface_"+dir(i)+".txt");
        LOG::fprintf(stream,"x y\n");
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        viewer_dir.Read_Last_Frame(0);

        GRID<TV> grid;
        ARRAY<T,IV> gamma;
        Read_From_File(viewer_dir.Common_File("grid"),grid);
        Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);

        //create surface and add to tex file
        MESH_OBJECT<TV,TV::m-1> surface;
        T contour_value = (T)0.5;
        MARCHING_CUBES<TV>::Create_Surface(surface,grid,at_cells,gamma,contour_value);
        surface.mesh.Initialize_Ordered_Components();
        bool is_surface=true;
        for(const auto &l:*surface.mesh.ordered_loop_nodes)
        {
            if(is_surface)
            {
                ARRAY<TV> pts(surface.particles.X.Subset(l));
                for (int j=0;j<pts.Size();++j) LOG::fprintf(stream,"%p %p\n",pts(j)(0),pts(j)(1));
            }
            is_surface=!is_surface;
        }
    }
}
//#####################################################################
// Function Compute_Perimeter
//#####################################################################
T Compute_Perimeter(const ARRAY<T,IV>& g,GRID<TV> grid,T contour)
{
    MESH_OBJECT<TV,TV::m-1> surface;
    MARCHING_CUBES<TV>::Create_Surface(surface,grid.Get_Regular_Grid_At_MAC_Positions(),at_nodes,g,contour);
    return surface.Total_Size();
}
//#####################################################################
// Function Computer_Area
//#####################################################################
T Compute_Area(const ARRAY<T,IV>& g,GRID<TV> grid,T contour)
{
    MESH_OBJECT<TV,TV::m> volume;
    MARCHING_CUBES<TV>::Create_Interior(volume,grid.Get_Regular_Grid_At_MAC_Positions(),at_nodes,g,true,contour);
    return volume.Total_Size();
}
//#####################################################################
// Function Dump_Circularity_data
//#####################################################################
void Dump_Circularity_data(const ARRAY<std::string> &dir)
{
    std::ofstream circularity_error(fig_dir+"/txt/circularity_data.txt");
    for(int i=0;i<dir.m;i++)
    {       
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        GRID<TV> grid;
        Read_From_File(viewer_dir.Common_File("grid"),grid);
        
        if(table)
        {
            LOG::cout<<"appending to circularity table file"<<std::endl;
            T res=grid.numbers_of_cells(0);
            viewer_dir.Read_Last_Frame(0);
            ARRAY<T,IV> gamma;
            Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);
            T perimeter=Compute_Perimeter(gamma,grid,0.5);
            T area=Compute_Area(gamma,grid,0.5);
            T circularity=2*sqrt(pi*area)/perimeter;
            circularity_error<<res<<" "<<circularity<<std::endl;
        }
        else
        {
            std::string filepath=fig_dir+"/txt/circularity_";
            if(ref) filepath+="reference_";
            filepath+=dir(i)+".txt";
            LOG::cout<<"creating file: "<<filepath<<std::endl;
            std::ofstream stream(filepath);
            LOG::fprintf(stream,"t circularity\n");
            int frame_num=0;
            bool not_last_frame=true;
            for(viewer_dir.Set(0);not_last_frame;not_last_frame=viewer_dir.Find_Next_Directory(0,true))
            {
                ARRAY<T,IV> gamma;
                Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);
                T perimeter=Compute_Perimeter(gamma,grid,0.5);
                T area=Compute_Area(gamma,grid,0.5);
                T circularity=2*sqrt(pi*area)/perimeter;
                T time=dt*frame_num;
                LOG::fprintf(stream,"%p %p\n",time,circularity);
                frame_num++;
            }
        }
    }
}
//#####################################################################
// Funcion Center_of_Mass
//#####################################################################
T Center_of_Mass(const GRID<TV>& grid,const ARRAY<T,IV>& gamma)
{
    T s=0,m=0;
    for(RANGE_ITERATOR<TV::m> it(grid.Cell_Indices());it.Valid();it.Next())
    {
        T g=gamma(it.index);
        T g_norm=1-(g-r)/(1-2*r);
        s+=grid.Cell_X(it.index)(1)*g_norm;
        m+=g_norm;
    }
    return s/m;
}
//#####################################################################
// Function Dump_COM_data
//#####################################################################
void Dump_COM_data(const ARRAY<std::string> &dir)
{
    std::ofstream COM_error(fig_dir+"/txt/COM_data.txt");
    for(int i=0;i<dir.m;i++)
    {
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        GRID <TV> grid;
        Read_From_File(viewer_dir.Common_File("grid"),grid);

        if(table)
        {
            LOG::cout<<"appending to COM table file"<<std::endl;
            T res=grid.numbers_of_cells(0);
            viewer_dir.Read_Last_Frame(0);
            ARRAY <T,IV> gamma;
            Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);
            T COM=Center_of_Mass(grid,gamma);
            COM_error<<res<<" "<<COM<<std::endl;
        }
        else
        {
            std::string filepath=fig_dir+"/txt/COM_";
            if(ref) filepath+="reference_";
            filepath+=dir(i)+".txt";
            LOG::cout<<"creating file: "<<filepath<<std::endl;
            std::ofstream stream(filepath);

            LOG::fprintf(stream, "t COM\n");

            bool not_last_frame=true;
            int frame_num=0;
            for(viewer_dir.Set(0);not_last_frame;not_last_frame=viewer_dir.Find_Next_Directory(0,true))
            {
                ARRAY <T,IV> gamma;
                Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);
                T COM=Center_of_Mass(grid,gamma);
                T time=dt*frame_num;
                LOG::fprintf(stream,"%p %p\n",time,COM);
                frame_num++;
            }
        }
    }
}
//#####################################################################
// Function Rising_Velocity
//#####################################################################
T Rising_Velocity(GRID<TV> grid,ARRAY<T,IV> gamma,ARRAY<T,FACE_INDEX<TV::m>> v)
{
    T s=0,m=0;
    AVERAGING_UNIFORM<TV> au;
    for(RANGE_ITERATOR<TV::m> it(grid.Cell_Indices());it.Valid();it.Next())
    {
        T g=gamma(it.index);
        s+=au.Face_To_Cell_Vector(grid,it.index,v).y*(1-(g-r)/(1-2*r));
        m+=(1-g);
    }
    return s/m;
}
//#####################################################################
// Function Dump_rising_V_data
//#####################################################################
void Dump_rising_V_data(const ARRAY<std::string> &dir)
{
    std::ofstream velocity_error(fig_dir+"/txt/velocity_data.txt");
    for(int i=0;i<dir.m;i++)
    {
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        GRID<TV> grid;
        Read_From_File(viewer_dir.Common_File("grid"),grid);
        if(table)
        {
            LOG::cout<<"appending to velocity table file"<<std::endl;
            T res=grid.numbers_of_cells(0);
            viewer_dir.Read_Last_Frame(0);
            ARRAY<T,IV> gamma;
            ARRAY<T,FACE_INDEX<TV::m>> v;
            Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);
            Read_From_File(viewer_dir.Frame_File("velocities"),v);
            T rising_v=Rising_Velocity(grid,gamma,v);
            velocity_error<<res<<" "<<rising_v<<std::endl;
        }
        else
        {
            std::string filepath=fig_dir+"/txt/velocity_";
            if(ref) filepath+="reference_";
            filepath+=dir(i)+".txt";
            LOG::cout<<"creating file: "<<filepath<<std::endl;
            std::ofstream stream(filepath);

            LOG::fprintf(stream,"t velocity\n");
            
            bool not_last_frame=true;
            int frame_num=0;
            for(viewer_dir.Set(0);not_last_frame;not_last_frame=viewer_dir.Find_Next_Directory(0,true))
            {
                ARRAY<T,IV> gamma;
                ARRAY<T,FACE_INDEX<TV::m>> v;
                Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);
                Read_From_File(viewer_dir.Frame_File("velocities"),v);
                T rising_v=Rising_Velocity(grid,gamma,v);
                T time=dt*frame_num;
                LOG::fprintf(stream,"%p %p\n",time,rising_v);
                frame_num++;
            }
        }
    }
}
//#####################################################################
// Function main
//#####################################################################
int main(int argc,char *argv[])
{
    PROCESS_UTILITIES::Set_Floating_Point_Exception_Handling(true);
    PROCESS_UTILITIES::Set_Backtrace(true);

    PARSE_ARGS parse_args(argc,argv);
    LOG::Initialize_Logging(false,false,1<<30,true);
    LOG::cout<<parse_args.Print_Arguments()<<std::endl;

    ARRAY<std::string> dir;
    parse_args.Add("-i",&dir,"string-directory name","directory name to read data from");
    parse_args.Add("-dt",&dt,"scalar","frame_dt");
    parse_args.Add("-table",&table,"create error table");
    parse_args.Add("-ref",&ref,"this is a reference sim");
    parse_args.Add("-ch_min",&r,"scalar","minimum gamma value");
    parse_args.Add("-fig",&fig_dir,"dir","base figure directory");

    parse_args.Parse();
    
    Dump_Circularity_data(dir);
    Dump_COM_data(dir);
    Dump_rising_V_data(dir);
    if(!ref&&!table) Dump_Surface_data(dir);
    
    LOG::Finish_Logging();
    return 0;
}
