#include <Core/Arrays/ARRAYS_ND.h>
#include <Core/Arrays/FACE_ARRAYS.h>
#include <Core/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Grids/GRID.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Utilities/PROCESS_UTILITIES.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Core/Viewer/VIEWER_OUTPUT.h>

using namespace PhysBAM;

typedef double T;
typedef VECTOR<T,2> TV;
typedef VECTOR<int,2> IV;

int main(int argc,char *argv[])
{
    PROCESS_UTILITIES::Set_Floating_Point_Exception_Handling(true);
    PROCESS_UTILITIES::Set_Backtrace(true);

    PARSE_ARGS parse_args(argc,argv);
    LOG::Initialize_Logging(false,false,1<<30,true);
    LOG::cout<<parse_args.Print_Arguments()<<std::endl;
    LOG::cout<<std::setprecision(16);

    std::string input1;
    std::string input2;
    std::string output="output";
    std::string density_name="phi";
    bool interleave=false;
    bool show_diff=false;
    parse_args.Add("-in1",&input1,"path","input directory 1");
    parse_args.Add("-in2",&input2,"path","input directory 2");
    parse_args.Add("-o",&output,"path","output directory");
    parse_args.Add("-i",&interleave,"interleave output frames");
    parse_args.Add("-diff",&show_diff,"show diff ");
    parse_args.Add("-d_name",&density_name,"path","name for density");
    parse_args.Parse(true);

    VIEWER_DIR vd_in1(input1);
    VIEWER_DIR vd_in2(input2);
    VIEWER_DIR vd_out(output);
    VIEWER_OUTPUT vo(STREAM_TYPE::type_double,vd_out);
    
    GRID<TV> grid1;
    GRID<TV> grid2;
    GRID<TV> grid_out;
    Read_From_File(vd_in1.Common_File("grid"),grid1);
    Read_From_File(vd_in2.Common_File("grid"),grid2);
    if(grid1.domain!=grid2.domain)
    {
        LOG::cout<<"Input grid domains are not the same."<<std::endl;
        return 1;
    }
    grid_out=grid1;
    vd_in1.Advance_Directory(0);
    vd_in2.Advance_Directory(0);

    ARRAY<T,IV> density1;
    ARRAY<T,IV> density2;
    ARRAY<T,IV> density_diff;
    ARRAY<T,IV> density_disp;

    ARRAY<T,FACE_INDEX<TV::m> > face_velocities1;
    ARRAY<T,FACE_INDEX<TV::m> > face_velocities2;
    ARRAY<T,FACE_INDEX<TV::m> > face_velocities_diff;
    ARRAY<T,FACE_INDEX<TV::m> > face_velocities_disp;

    vo.Add_Global({.is_2d=true});
    vo.Add("debug_particles",DEBUG_PARTICLES<TV>::Get(),false,{});
    vo.Add("grid",grid_out,true,{});
    vo.Add("density","grid",density_disp,false,{.cell_centered=true});
        // viewer_output.Add("temperature","grid",phi,false,{cell_centered=true});
    vo.Add("mac_velocities","grid",face_velocities_disp,false,{.color={{0,1,0}}});
        // viewer_output.Add("centered_velocities","grid",display_velocities_cell,false,{.color={{1,0,1}}});
        // viewer_output.Add("pressure","grid",pressure,false,{.cell_centered=true});

    int count=0;
    while(vd_in1.Valid_Frame(true)&&vd_in2.Valid_Frame(true))
    {
        // LOG::cout<<"Comparing frames \""<<vd_in1.current_directory<<"\" and \""<<vd_in2.current_directory<<"\""<<std::endl;
        LOG::cout<<"Comparing frame "<<count<<": "<<std::endl;
        Read_From_File(vd_in1.Frame_File("density"),density1);
        Read_From_File(vd_in2.Frame_File("density"),density2);
        density_diff.Resize(density1.domain);

        T dL2=0;
        T dc=0;
        for(CELL_ITERATOR<TV> it(grid_out);it.Valid();it.Next())
        {
            T error=density1(it.index)-density2(it.index);
            density_diff(it.index)=error;
            Add_Ball(grid_out.Cell_X(it.index),error,VECTOR<T,3>(1,0,0));
            dL2+=sqr(error);
            dc++;
        }
        if(dc) dL2=sqrt(dL2/dc);
        LOG::cout<<"\t "<<density_name<<" L2: "<<dL2<<std::endl;

        Read_From_File(vd_in1.Frame_File("velocities"),face_velocities1);
        Read_From_File(vd_in2.Frame_File("mac_velocities"),face_velocities2);
        face_velocities_diff.Resize(face_velocities1.domain);

        TV fvL2;
        IV fvc;
        for(FACE_ITERATOR<TV> it(grid_out);it.Valid();it.Next())
        {
            T error=face_velocities1(it.face)-face_velocities2(it.face);
            face_velocities_diff(it.face)=error;
            Add_Arrow(grid_out.Face_X(it.face),TV().Add_Axis(it.face.axis,error),VECTOR<T,3>(1,0.5,0));
            fvL2(it.face.axis)+=sqr(error);
            fvc(it.face.axis)++;
        }
        if(fvc.Magnitude()) fvL2=sqrt(fvL2/TV(fvc));
        LOG::cout<<"\t face velocity L2 (x): "<<fvL2(0)<<std::endl;
        LOG::cout<<"\t face velocity L2 (y): "<<fvL2(1)<<std::endl;
        
        
        density_disp.Exchange(density1);
        face_velocities_disp.Exchange(face_velocities1);
        std::string frame_name="Frame "+std::to_string(count)+": input 1";
        Flush_Frame(frame_name.c_str());
        
        if(interleave)
        {
            for(CELL_ITERATOR<TV> it(grid_out);it.Valid();it.Next())
                Add_Ball(grid_out.Cell_X(it.index),density_diff(it.index),VECTOR<T,3>(1,0,0));
            for(FACE_ITERATOR<TV> it(grid_out);it.Valid();it.Next())
                Add_Arrow(grid_out.Face_X(it.face),TV().Add_Axis(it.face.axis,face_velocities_diff(it.face)),VECTOR<T,3>(1,0.5,0));
            
            density_disp.Exchange(density2);
            face_velocities_disp.Exchange(face_velocities2);
            frame_name="Frame "+std::to_string(count)+": input 2";
            Flush_Frame(frame_name.c_str());
        }

        if(show_diff)
        {
            for(CELL_ITERATOR<TV> it(grid_out);it.Valid();it.Next())
                Add_Ball(grid_out.Cell_X(it.index),density_diff(it.index),VECTOR<T,3>(1,0,0));
            for(FACE_ITERATOR<TV> it(grid_out);it.Valid();it.Next())
                Add_Arrow(grid_out.Face_X(it.face),TV().Add_Axis(it.face.axis,face_velocities_diff(it.face)),VECTOR<T,3>(1,0.5,0));
            density_disp.Exchange(density_diff);
            face_velocities_disp.Exchange(face_velocities_diff);
            frame_name="Frame "+std::to_string(count)+": input diff";
            Flush_Frame(frame_name.c_str());
        }
        count++;
        
        vd_in1.Advance_Directory(0);
        vd_in2.Advance_Directory(0);
    }
    
    LOG::Finish_Logging();
    return 0;
}
