//##############################################################################
// August 2023 - Mehrnaz Ayazi, Jason Goulding
//##############################################################################
#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays/FACE_ARRAYS.h>
#include <Core/Grids/FACE_ITERATOR.h>
#include <Core/Grids/GRID.h>
#include <Core/Math_Tools/RANGE.h>
#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Utilities/PROCESS_UTILITIES.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <fstream>

using namespace PhysBAM;

typedef double T;
typedef VECTOR<T,2> TV;
typedef VECTOR<int,2> IV;

std::string fig_dir;
ARRAY<std::string> dir;
ARRAY<T> dt;
T rho0=1,rho1=1;
VECTOR<bool,2> periodic={true,true};
//#####################################################################
// Function Dump_Mass_Change_data
//#####################################################################
void Dump_Mass_Change_data()
{
    for(int i=0;i<dir.m;i++)
    {
        std::string filename="masschange_"+dir(i)+".txt";
        LOG::cout<<"creating data file "<<fig_dir+"/txt/"+filename<<std::endl;
        std::ofstream stream(fig_dir+"/txt/"+filename);
        
        LOG::fprintf(stream,"t masschange\n");
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        GRID<TV> grid;
        Read_From_File(viewer_dir.Common_File("grid"),grid);
        bool not_last_frame=true;
        int frame_num=0;
        T time=0;
        T initial_total_mass=0;
        viewer_dir.Set(0);
        {
            ARRAY<T,IV> gamma;
            Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);
            ARRAY<T,IV> mass(gamma.domain);
            mass.array=(rho0+(rho1-rho0)*gamma.array)*grid.Cell_Size();
            T total_mass=mass.Sum();
            initial_total_mass=total_mass;
            time=dt(i)*frame_num;
            LOG::fprintf(stream,"%P %.16P\n",time,total_mass-initial_total_mass);
            frame_num++;
        }
        for(not_last_frame=viewer_dir.Find_Next_Directory(0,true);not_last_frame;not_last_frame=viewer_dir.Find_Next_Directory(0,true))
        {
            ARRAY<T,IV> gamma;
            Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);
            ARRAY<T,IV> mass(gamma.domain);
            mass.array=(rho0+(rho1-rho0)*gamma.array)*grid.Cell_Size();
            T total_mass=mass.Sum();
            time=dt(i)*frame_num;
            LOG::fprintf(stream,"%P %.16P\n",time,total_mass-initial_total_mass);
            frame_num++;
        }
        // LOG::cout<<"frame num="<<frame_num<<std::endl;
    }
}
//#####################################################################
// Function Total_KE
//#####################################################################
T Total_KE(const GRID<TV>& grid,const ARRAY<T,FACE_INDEX<TV::m> >& u,const ARRAY<T,FACE_INDEX<TV::m> >& density)
{
    T ke=0;
    for(FACE_ITERATOR<TV> it(grid,0,INTERIOR_REGION);it.Valid();it.Next())
        ke+=density(it.face)*sqr(u(it.face))*grid.Cell_Size()/2;
    
    for(int axis=0;axis<TV::m;axis++)
    {
        if(periodic(axis))
            for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,2*axis);it.Valid();it.Next())
                ke+=density(it.face)*sqr(u(it.face))*grid.Cell_Size()/2;
        else
            for(int side=0;side<2;side++)
                for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,2*axis+side);it.Valid();it.Next())
                    ke+=density(it.face)*sqr(u(it.face))*grid.Cell_Size()/4;
    }
    return ke;
}
//#####################################################################
// Function Dump_KE_Change_data
//#####################################################################
void Dump_KE_Change_data()
{
    for(int i=0;i<dir.m;i++)
    {
        std::string filename="kechange_"+dir(i)+".txt";
        LOG::cout<<"creating data file "<<fig_dir+"/txt/"+filename<<std::endl;
        std::ofstream stream(fig_dir+"/txt/"+filename);
        
        LOG::fprintf(stream,"t kechange\n");
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        GRID<TV> grid;
        Read_From_File(viewer_dir.Common_File("grid"),grid);
        bool not_last_frame=true;
        int frame_num=0;
        T time=0;
        T initial_total_ke=0;
        viewer_dir.Set(0);
        {
            ARRAY<T,FACE_INDEX<TV::m> > velocities;
            ARRAY<T,FACE_INDEX<TV::m> > density;
            Read_From_File(viewer_dir.Frame_File("face_velocities"),velocities);
            Read_From_File(viewer_dir.Frame_File("face_density"),density);

            T total_ke=Total_KE(grid,velocities,density);
            initial_total_ke=total_ke;
            time=dt(i)*frame_num;
            LOG::fprintf(stream,"%P %.16P\n",time,total_ke-initial_total_ke);
            frame_num++;
        }
        for(not_last_frame=viewer_dir.Find_Next_Directory(0,true);not_last_frame;not_last_frame=viewer_dir.Find_Next_Directory(0,true))
        {
            ARRAY<T,FACE_INDEX<TV::m> > velocities;
            ARRAY<T,FACE_INDEX<TV::m> > density;
            Read_From_File(viewer_dir.Frame_File("face_velocities"),velocities);
            Read_From_File(viewer_dir.Frame_File("face_density"),density);

            T total_ke=Total_KE(grid,velocities,density);
            time=dt(i)*frame_num;
            LOG::fprintf(stream,"%P %.16P\n",time,total_ke-initial_total_ke);
            frame_num++;
        }
        // LOG::cout<<"frame num="<<frame_num<<std::endl;
    }
}
//#####################################################################
// Function Total_Momentum
//#####################################################################
TV Total_Momentum(const GRID<TV>& grid,const ARRAY<T,FACE_INDEX<TV::m> >& u,const ARRAY<T,FACE_INDEX<TV::m> >& density)
{    
    TV momentum;

    for(FACE_ITERATOR<TV> it(grid,0,INTERIOR_REGION);it.Valid();it.Next())
        momentum(it.face.axis)+=density(it.face)*u(it.face)*grid.Cell_Size();
    
    for(int axis=0;axis<TV::m;axis++)
    {
        if(periodic(axis))
            for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,2*axis);it.Valid();it.Next())
                momentum(it.face.axis)+=density(it.face)*u(it.face)*grid.Cell_Size();
        else
            for(int side=0;side<2;side++)
                for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION,2*axis+side);it.Valid();it.Next())
                    momentum(it.face.axis)+=density(it.face)*u(it.face)*grid.Cell_Size()/2;
    }
    return momentum;
}
//#####################################################################
// Function Dump_Momentum_Change_data
//#####################################################################
void Dump_Momentum_Change_data()
{
    for(int i=0;i<dir.m;i++)
    {
        std::string filename_x="mochangex_"+dir(i)+".txt";
        LOG::cout<<"creating data file "<<fig_dir+"/txt/"+filename_x<<std::endl;
        std::ofstream stream_x(fig_dir+"/txt/"+filename_x);
        std::string filename_y="mochangey_"+dir(i)+".txt";
        LOG::cout<<"creating data file "<<fig_dir+"/txt/"+filename_y<<std::endl;
        std::ofstream stream_y(fig_dir+"/txt/"+filename_y);
        
        LOG::fprintf(stream_x,"t mochangex\n");
        LOG::fprintf(stream_y,"t mochangey\n");
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        GRID<TV> grid;
        Read_From_File(viewer_dir.Common_File("grid"),grid);
        bool not_last_frame=true;
        int frame_num=0;
        T time=0;
        TV initial_total_momentum;
        viewer_dir.Set(0);
        {
            ARRAY<T,FACE_INDEX<TV::m> > velocities;
            ARRAY<T,FACE_INDEX<TV::m> > density;
            Read_From_File(viewer_dir.Frame_File("face_velocities"),velocities);
            Read_From_File(viewer_dir.Frame_File("face_density"),density);

            TV total_momentum=Total_Momentum(grid,velocities,density);
            initial_total_momentum=total_momentum;
            time=dt(i)*frame_num;
            LOG::fprintf(stream_x,"%P %.16P\n",time,total_momentum.x-initial_total_momentum.x);
            LOG::fprintf(stream_y,"%P %.16P\n",time,total_momentum.y-initial_total_momentum.y);
            frame_num++;
        }
        for(not_last_frame=viewer_dir.Find_Next_Directory(0,true);not_last_frame;not_last_frame=viewer_dir.Find_Next_Directory(0,true))
        {
            ARRAY<T,FACE_INDEX<TV::m> > velocities;
            ARRAY<T,FACE_INDEX<TV::m> > density;
            Read_From_File(viewer_dir.Frame_File("face_velocities"),velocities);
            Read_From_File(viewer_dir.Frame_File("face_density"),density);

            TV total_momentum=Total_Momentum(grid,velocities,density);
            time=dt(i)*frame_num;
            LOG::fprintf(stream_x,"%P %.16P\n",time,total_momentum.x-initial_total_momentum.x);
            LOG::fprintf(stream_y,"%P %.16P\n",time,total_momentum.y-initial_total_momentum.y);
            frame_num++;
        }
        // LOG::cout<<"frame num="<<frame_num<<std::endl;
    }
}
//#####################################################################
// Function main
//#####################################################################
int main(int argc,char *argv[])
{
    dt.Fill(1.0/24);
    PROCESS_UTILITIES::Set_Floating_Point_Exception_Handling(true);
    PROCESS_UTILITIES::Set_Backtrace(true);

    PARSE_ARGS parse_args(argc,argv);
    // LOG::Initialize_Logging(false,false,1<<30,true);
    LOG::cout<<parse_args.Print_Arguments()<<std::endl;

    VECTOR<T,2> range_scale={0,1};
    std::string boundary_file="";

    parse_args.Add("-dir",&dir,"string-array","directory name to read data from");
    parse_args.Add("-dt",&dt,"scalar-array","frame_dt");
    parse_args.Add("-rho0",&rho0,"scalar","density 0");
    parse_args.Add("-rho1",&rho1,"scalar","density 1");
    parse_args.Add("-periodic",&periodic,"bool vector","Define the axii where the domain should be periodic");
    parse_args.Add("-fig",&fig_dir,"dir","base figure directory");

    parse_args.Parse();
    Dump_Mass_Change_data();
    Dump_KE_Change_data();
    Dump_Momentum_Change_data();
    // LOG::Finish_Logging();
    return 0;
}
