//##############################################################################
// MARCH2023-Mehrnaz Ayazi
//##############################################################################
#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays/FACE_ARRAYS.h>
#include <Core/Grids/GRID.h>
#include <Core/Grids_Uniform_Computations/MARCHING_CUBES.h>
#include <Core/Images/TEX_FILE.h>
#include <Core/Interpolation/AVERAGING_UNIFORM.h>
#include <Core/Math_Tools/RANGE.h>
#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Core/Meshes/MESH_OBJECT.h>
#include <Core/Meshes/SEGMENTED_CURVE.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Utilities/PROCESS_UTILITIES.h>
#include <Core/Utilities/VIEWER_DIR.h>


using namespace PhysBAM;

typedef double T;
typedef VECTOR<T,2> TV;
typedef VECTOR<int,2> IV;
typedef VECTOR<T,3> COLOR;

ARRAY<COLOR> colors;
ARRAY<std::string> dir;
ARRAY<T> dt;
ARRAY<T> res;
T rho0=1,rho1=1;
T r=0;

std::string fig_dir;
std::ofstream circularity_error;
std::ofstream COM_error;
std::ofstream velocity_error;
bool table=false;
bool ref=false;
//#####################################################################
// Function Dump_Surface_data
//#####################################################################
void Dump_Surface_data(const std::string& file)
{
    for(int i=0;i<dir.m;i++)
    {
        std::ofstream stream(fig_dir+"/txt/"+dir(i)+"_"+file);
        LOG::fprintf(stream,"x y\n");
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        viewer_dir.Read_Last_Frame(0);

        GRID<TV> grid;
        ARRAY<T,IV> gamma;
        Read_From_File(viewer_dir.Common_File("grid"),grid);
        Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);

        //create surface and add to tex file
        MESH_OBJECT<TV,TV::m-1> surface;
        T contour_value = (T)0.5;
        MARCHING_CUBES<TV>::Create_Surface(surface,grid,at_cells,gamma,contour_value);
        surface.mesh.Initialize_Ordered_Components();
        bool is_surface=true;
        for(const auto &l:*surface.mesh.ordered_loop_nodes)
        {
            if(is_surface){
                ARRAY<TV> pts(surface.particles.X.Subset(l));
                for (int j=0;j<pts.Size();++j) {
                    LOG::fprintf(stream,"%p %p\n",pts(j)(0),pts(j)(1));
                }
            }
            is_surface=!is_surface;
        }
    }
}
//#####################################################################
// Function Dump_Surface
//#####################################################################
void Dump_Surface(const std::string& file)
{
    TEX_FILE<T> tex(file);
    tex.cur_format.fill_style=0;
    for(int i=0;i<dir.m;i++)
    {
        tex.cur_format.line_color=colors(i%colors.m);
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        viewer_dir.Read_Last_Frame(0);

        GRID<TV> grid;
        ARRAY<T,IV> gamma;
        Read_From_File(viewer_dir.Common_File("grid"),grid);
        Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);

        //create surface and add to tex file
        MESH_OBJECT<TV,TV::m-1> surface;
        T contour_value=(T)0.5;
        MARCHING_CUBES<TV>::Create_Surface(surface,grid,at_cells,gamma,contour_value);
        surface.mesh.Initialize_Ordered_Components();
        for(const auto &l:*surface.mesh.ordered_loop_nodes)
        {
            ARRAY<TV> pts(surface.particles.X.Subset(l));
            tex.Draw_Object(pts);
        }
    }
    tex.cur_format.line_color={1,0,0};
    RANGE<TV> box(tex.bounding_box);
    box.Scale_About_Center(1.1);
    tex.Draw_Object(box);
}
//#####################################################################
// Function Compute_Perimeter
//#####################################################################
T Compute_Perimeter(const ARRAY<T,IV>& g,GRID<TV> grid, T contour)
{
    MESH_OBJECT<TV,TV::m-1> surface;
    MARCHING_CUBES<TV>::Create_Surface(surface,grid.Get_Regular_Grid_At_MAC_Positions(),at_nodes,g,contour);
    return surface.Total_Size();
}
//#####################################################################
// Function Computer_Area
//#####################################################################
T Compute_Area(const ARRAY<T,IV>& g,GRID<TV> grid, T contour)
{
    MESH_OBJECT<TV,TV::m> volume;
    MARCHING_CUBES<TV>::Create_Interior(volume,grid.Get_Regular_Grid_At_MAC_Positions(),at_nodes,g,true,contour);
    return volume.Total_Size();
}
//#####################################################################
// Function Dump_Circularity_data
//#####################################################################
void Dump_Circularity_data(const std::string& file)
{
    for(int i=0;i<dir.m;i++)
    {
        std::string filename;
        if(ref){filename=fig_dir+"/txt/"+file;LOG::cout<<"creating reference file "<<filename<<std::endl;}
        else{filename=fig_dir+"/txt/"+dir(i)+"_"+file;LOG::cout<<"creating data file "<<filename<<std::endl;}
        std::ofstream stream(filename);
        LOG::fprintf(stream,"t circularity\n");
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        GRID<TV> grid;
        Read_From_File(viewer_dir.Common_File("grid"),grid);
        bool not_last_frame = true;
        int frame_num = 0;
        T time=0;
        T perimeter,area,circularity;
        for (viewer_dir.Set(0);not_last_frame;not_last_frame=viewer_dir.Find_Next_Directory(0,true))
        {
            ARRAY<T,IV> gamma;
            Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);
            perimeter=Compute_Perimeter(gamma,grid,0.5);
            area=Compute_Area(gamma,grid,0.5);
            circularity=2*sqrt(pi*area)/perimeter;
            time=dt(i)*frame_num;
            LOG::fprintf(stream,"%p %p\n",time,circularity);
            frame_num++;
        }
        LOG::cout<<"frame num="<<frame_num<<std::endl;

        if(table){
            viewer_dir.Read_Last_Frame(0);
            ARRAY<T,IV> gamma;
            Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);
            perimeter=Compute_Perimeter(gamma,grid,0.5);
            area=Compute_Area(gamma,grid,0.5);
            circularity=2*sqrt(pi*area)/perimeter;
            circularity_error<<res(i)<<" "<<circularity<<std::endl;
        }
    }
}
//#####################################################################
// Function Dump_Circularity_graph
//#####################################################################
void Dump_Circularity_graph(const std::string& file)
{
    TEX_FILE<T> tex(file);
    tex.cur_format.fill_style=0;
    for(int i=0;i<dir.m;i++)
    {
        tex.cur_format.line_color = colors(i%colors.m);
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        GRID<TV> grid;
        Read_From_File(viewer_dir.Common_File("grid"),grid);

        bool not_last_frame = true;
        int frame_num = 0;
        TV a,b;
        T time=0;
        T perimeter,area,circularity;
        for (viewer_dir.Set(0);not_last_frame;not_last_frame=viewer_dir.Find_Next_Directory(0,true))
        {
            a=b;
            ARRAY<T,IV> gamma;
            Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);
            perimeter=Compute_Perimeter(gamma,grid,0.5);
            area=Compute_Area(gamma,grid,0.5);
            circularity=2*sqrt(pi*area)/perimeter;
            time=dt(i)*frame_num;
            b={time,circularity};
            if(frame_num>0)
                tex.Draw_Object(a,b);
            frame_num++;
        }
    }
    tex.cur_format.line_color={1,0,0};
    RANGE<TV> box(tex.bounding_box);
    box.Scale_About_Center(1.1);
    tex.Draw_Object(box);
}
//#####################################################################
// Funcion Center_of_Mass
//#####################################################################
T Center_of_Mass(const GRID<TV>& grid,const ARRAY<T,IV>& gamma)
{
    T s=0,m=0;
    for(RANGE_ITERATOR<TV::m> it(grid.Cell_Indices());it.Valid();it.Next())
    {
        T g=gamma(it.index);
        s+=grid.Cell_X(it.index)(1)*(1-(g-r)/(1-2*r));
        m+=(1-g);
    }
    return s/m;
}
//#####################################################################
// Function Dump_COM_data
//#####################################################################
void Dump_COM_data(const std::string& file)
{
    for(int i=0;i<dir.m;i++) {
        std::ofstream stream;
        if (ref) stream.open(fig_dir+"/txt/" + file);
        else stream.open(fig_dir+"/txt/" + dir(i) + "_" + file);
        LOG::fprintf(stream, "t COM\n");
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        GRID <TV> grid;
        Read_From_File(viewer_dir.Common_File("grid"), grid);

        bool not_last_frame = true;
        int frame_num = 0;
        T time = 0;
        T COM;
        for (viewer_dir.Set(0); not_last_frame; not_last_frame = viewer_dir.Find_Next_Directory(0, true)) {
            ARRAY <T, IV> gamma;
            Read_From_File(viewer_dir.Frame_File("display_gamma"), gamma);
            COM = Center_of_Mass(grid, gamma);
            time = dt(i) * frame_num;
            LOG::fprintf(stream, "%p %p\n", time, COM);
            frame_num++;
        }
        if (table) {
            viewer_dir.Read_Last_Frame(0);
            ARRAY <T, IV> gamma;
            Read_From_File(viewer_dir.Frame_File("display_gamma"), gamma);
            COM = Center_of_Mass(grid, gamma);
            COM_error << res(i) << " " << COM << std::endl;
        }
    }
}
//#####################################################################
// Function Dump_COM_graph
//#####################################################################
void Dump_COM_graph(const std::string& file)
{
    TEX_FILE<T> tex(file);
    tex.cur_format.fill_style=0;
    for(int i=0;i<dir.m;i++)
    {
        tex.cur_format.line_color = colors(i%colors.m);
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        GRID<TV> grid;
        Read_From_File(viewer_dir.Common_File("grid"),grid);

        bool not_last_frame = true;
        int frame_num = 0;
        TV a,b;
        T time=0;
        T COM;
        for (viewer_dir.Set(0);not_last_frame;not_last_frame = viewer_dir.Find_Next_Directory(0,true))
        {
            a=b;
            ARRAY<T,IV> gamma;
            Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);
            COM=Center_of_Mass(grid,gamma);
            time=dt(i)*frame_num;
            time=dt(i)*frame_num;
            b={time,COM};
            LOG::cout<<b<<std::endl;
            if(frame_num>0)
                tex.Draw_Object(a,b);
            frame_num++;
        }
    }
    tex.cur_format.line_color={1,0,0};
    RANGE<TV> box(tex.bounding_box);
    box.Scale_About_Center(1.1);
    tex.Draw_Object(box);
}
//#####################################################################
// Function Rising_Velocity
//#####################################################################
T Rising_Velocity(GRID<TV> grid,ARRAY<T,IV> gamma,ARRAY<T,FACE_INDEX<TV::m>> v)
{
    T s=0,m=0;
    AVERAGING_UNIFORM<TV> au;
    for(RANGE_ITERATOR<TV::m> it(grid.Cell_Indices());it.Valid();it.Next())
    {
        T g=gamma(it.index);
        s+=au.Face_To_Cell_Vector(grid,it.index,v).y*(1-(g-r)/(1-2*r));
        m+=(1-g);
    }
    return s/m;
}
//#####################################################################
// Function Dump_rising_V_data
//#####################################################################
void Dump_rising_V_data(const std::string& file)
{
    for(int i=0;i<dir.m;i++)
    {
        std::ofstream stream;
        if(ref) {stream.open(fig_dir+"/txt/"+file);LOG::cout<<"opening file:"<<fig_dir+"/txt/"+file<<std::endl;}
        else stream.open(fig_dir+"/txt/"+dir(i)+"_"+file);
        LOG::fprintf(stream,"t velocity\n");
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        GRID<TV> grid;
        Read_From_File(viewer_dir.Common_File("grid"),grid);

        bool not_last_frame = true;
        int frame_num = 0;
        T time=0;
        T rising_v;
        for (viewer_dir.Set(0);not_last_frame;not_last_frame = viewer_dir.Find_Next_Directory(0,true)) {
            ARRAY<T,IV> gamma;
            ARRAY<T,FACE_INDEX<TV::m>> v;
            Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);
            Read_From_File(viewer_dir.Frame_File("velocities"),v);
            rising_v=Rising_Velocity(grid,gamma,v);
            time=dt(i)*frame_num;
            LOG::fprintf(stream,"%p %p\n",time,rising_v);
            frame_num++;
        }
        if(table){
            viewer_dir.Read_Last_Frame(0);
            ARRAY<T,IV> gamma;
            ARRAY<T,FACE_INDEX<TV::m>> v;
            Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);
            Read_From_File(viewer_dir.Frame_File("velocities"),v);
            rising_v=Rising_Velocity(grid,gamma,v);
            velocity_error<<res(i)<<" "<<rising_v<<std::endl;
        }
    }
}
//#####################################################################
// Function Dump_rising_V_graph
//#####################################################################
void Dump_rising_V_graph(const std::string& file)
{
    TEX_FILE<T> tex(file);
    tex.cur_format.fill_style=0;
    for(int i=0;i<dir.m;i++)
    {
        tex.cur_format.line_color = colors(i%colors.m);
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        GRID<TV> grid;
        Read_From_File(viewer_dir.Common_File("grid"),grid);

        bool not_last_frame = true;
        int frame_num = 0;
        TV a,b;
        T time=0;
        T rising_v;
        for (viewer_dir.Set(0);not_last_frame;not_last_frame = viewer_dir.Find_Next_Directory(0,true)) {
            a=b;
            ARRAY<T,IV> gamma;
            ARRAY<T,FACE_INDEX<TV::m>> v;
            Read_From_File(viewer_dir.Frame_File("display_gamma"),gamma);
            Read_From_File(viewer_dir.Frame_File("velocities"),v);
            rising_v=Rising_Velocity(grid,gamma,v);
            time=dt(i)*frame_num;
            b={time,rising_v};
            if(frame_num>0)
                tex.Draw_Object(a,b);
            frame_num++;
        }
    }
    tex.cur_format.line_color={1,0,0};
    RANGE<TV> box(tex.bounding_box);
    box.Scale_About_Center(1.1);
    tex.Draw_Object(box);
}

//#####################################################################
// Function main
//#####################################################################
int main(int argc,char *argv[])
{
    dt.Fill(1/24);
    PROCESS_UTILITIES::Set_Floating_Point_Exception_Handling(true);
    PROCESS_UTILITIES::Set_Backtrace(true);

    PARSE_ARGS parse_args(argc,argv);
    LOG::Initialize_Logging(false,false,1<<30,true);
    LOG::cout<<parse_args.Print_Arguments()<<std::endl;

    VECTOR<T,2> range_scale={0,1};
    std::string boundary_file="";

    parse_args.Add("-ob",&boundary_file,"string-file name","file name for data");
    parse_args.Add("-i",&dir,"string-directory name","directory name to read data from");
    parse_args.Add("-c",&colors,"string-color name","line color for -make_fig");
    parse_args.Add("-dt",&dt,"scalar-array","frame_dt");
    parse_args.Add("-r0",&rho0,"scalar","density 0");
    parse_args.Add("-r1",&rho1,"scalar","density 1");
    parse_args.Add("-res",&res,"scalar-array","resolution");
    parse_args.Add("-table",&table,"create error table");
    parse_args.Add("-ref",&ref,"this is a reference sim");
    parse_args.Add("-ch_min",&r,"scalar","minimum gamma value");
    parse_args.Add("-fig",&fig_dir,"dir","base figure directory");

    parse_args.Parse();
    LOG::cout<<"ref="<<ref<<std::endl;

    circularity_error.open(fig_dir+"/txt/circularity_data.txt");
    COM_error.open(fig_dir+"/txt/COM_data.txt");
    velocity_error.open(fig_dir+"/txt/velocity_data.txt");

    std::string boundary_file_tex=boundary_file+".tex";
    std::string boundary_file_text=boundary_file+".txt";
    Dump_Circularity_data("circularity_"+boundary_file_text);
    Dump_COM_data("COM_"+boundary_file_text);
    Dump_rising_V_data("velocity_"+boundary_file_text);
    Dump_Surface_data("surface_"+boundary_file_text);
    LOG::Finish_Logging();
    return 0;
}
