//##############################################################################
// Jun 2023-Mehrnaz Ayazi
//##############################################################################
#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays/FACE_ARRAYS.h>
#include <Core/Grids/CELL_ITERATOR.h>
#include <Core/Grids/GRID.h>
#include <Core/Grids_Uniform_Computations/MARCHING_CUBES.h>
#include <Core/Images/TEX_FILE.h>
#include <Core/Interpolation/AVERAGING_UNIFORM.h>
#include <Core/Math_Tools/RANGE.h>
#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Core/Meshes/MESH_OBJECT.h>
#include <Core/Meshes/SEGMENTED_CURVE.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Utilities/PROCESS_UTILITIES.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <cmath>

using namespace PhysBAM;

typedef double T;
typedef VECTOR<T,2> TV;
typedef VECTOR<int,2> IV;
typedef VECTOR<T,3> COLOR;

ARRAY<COLOR> colors;
std::string fig_dir;
ARRAY<std::string> dir;
ARRAY<int> res;
ARRAY<T> dt;
ARRAY<T> sl;
bool cycle=false;
T rho0=0,rho1=1;
//#####################################################################
// Function Dump_Surface_data
//#####################################################################
void Dump_Surface_data(const std::string& file)
{
    std::ofstream stream2(fig_dir+"/txt/L2_"+file);
    std::ofstream streaminf(fig_dir+"/txt/Linf_"+file);
    LOG::fprintf(stream2,"res error\n");
    LOG::fprintf(streaminf,"res error\n");

    for(int i=0;i<dir.m;i++)
    {
        VIEWER_DIR viewer_dir(fig_dir+"/"+dir(i));
        viewer_dir.Read_Last_Frame(0);

        GRID<TV> grid;
        Read_From_File(viewer_dir.Common_File("grid"),grid);

        ARRAY<T,FACE_INDEX<TV::m> > face_velocities;
        Read_From_File(viewer_dir.Frame_File("velocities"),face_velocities);

        AVERAGING_UNIFORM<TV> avg;

        T L2_cell=0;
        T Linf_cell=0;
        int count=0;
        for(CELL_ITERATOR<TV> it(grid);it.Valid();it.Next())
        {
            TV v_cell=avg.Face_To_Cell_Vector(grid,it.index,face_velocities);
            T local=v_cell.Magnitude();
            Linf_cell=max(Linf_cell,local);
            L2_cell+=sqr(local);
            count++;
        }
        L2_cell/=count;
        LOG::fprintf(stream2,"%p %p\n",(res(i)*16),(sqrt(L2_cell)));
        LOG::fprintf(streaminf,"%p %p\n",(res(i)*16),(Linf_cell));
    }
}
//#####################################################################
// Function Dump_Reference_data
//#####################################################################
void Dump_Reference_data(T L2scale,T Linfscale)
{
    std::ofstream stream2_1(fig_dir+"/txt/L2_error_test_case6.txt");//(D/h)^-1
    std::ofstream stream2_2(fig_dir+"/txt/L2_error_test_case7.txt");//(D/h)^-2
    std::ofstream streaminf_1(fig_dir+"/txt/Linf_error_test_case6.txt");//(D/h)^-1
    std::ofstream streaminf_2(fig_dir+"/txt/Linf_error_test_case7.txt");//(D/h)^-2

    std::ofstream stream_const2_1(fig_dir+"/txt/L2_error_test_constant_case6.txt");//(D/h)^-1
    std::ofstream stream_const2_2(fig_dir+"/txt/L2_error_test_constant_case7.txt");//(D/h)^-2
    std::ofstream stream_constinf_1(fig_dir+"/txt/Linf_error_test_constant_case6.txt");//(D/h)^-1
    std::ofstream stream_constinf_2(fig_dir+"/txt/Linf_error_test_constant_case7.txt");//(D/h)^-2


    LOG::fprintf(stream2_1,"res error\n");
    LOG::fprintf(stream2_2,"res error\n");
    LOG::fprintf(streaminf_1,"res error\n");
    LOG::fprintf(streaminf_2,"res error\n");

    LOG::fprintf(stream_const2_1,"res error\n");
    LOG::fprintf(stream_const2_2,"res error\n");
    LOG::fprintf(stream_constinf_1,"res error\n");
    LOG::fprintf(stream_constinf_2,"res error\n");


    for(int i=0;i<dir.m;i++)
    {
        T real_res=res(i)*16;
        LOG::fprintf(stream2_1,"%p %p\n",(res(i)*16),(L2scale*0.4/real_res));
        LOG::fprintf(stream2_2,"%p %p\n",(res(i)*16),(L2scale*0.16/(real_res*real_res)));
        LOG::fprintf(streaminf_1,"%p %p\n",(res(i)*16),(Linfscale*0.4/real_res));
        LOG::fprintf(streaminf_2,"%p %p\n",(res(i)*16),(Linfscale*0.16/(real_res*real_res)));

        LOG::fprintf(stream_const2_1,"%p %p\n",(res(i)*16),(L2scale*0.4/real_res));
        LOG::fprintf(stream_const2_2,"%p %p\n",(res(i)*16),(L2scale*0.16/(real_res*real_res)));
        LOG::fprintf(stream_constinf_1,"%p %p\n",(res(i)*16),(Linfscale*0.4/real_res));
        LOG::fprintf(stream_constinf_2,"%p %p\n",(res(i)*16),(Linfscale*0.16/(real_res*real_res)));


    }
}
//#####################################################################
// Function main
//#####################################################################
int main(int argc,char *argv[])
{
    dt.Fill(1/24);
    PROCESS_UTILITIES::Set_Floating_Point_Exception_Handling(true);
    PROCESS_UTILITIES::Set_Backtrace(true);

    PARSE_ARGS parse_args(argc,argv);
    LOG::Initialize_Logging(false,false,1<<30,true);
    LOG::cout<<parse_args.Print_Arguments()<<std::endl;

    VECTOR<T,2> range_scale={0,1};
    std::string boundary_file="";
    T L2scale=1;
    T Linfscale=1;

    parse_args.Add("-ob",&boundary_file,"string-file name","file name for data");
    parse_args.Add("-i",&dir,"string-directory name","directory name to read data from");
    parse_args.Add("-res",&res,"int-resolution","resolution of sim");
    parse_args.Add("-L2_scale",&L2scale,"scalar","scale the L2 reference line");
    parse_args.Add("-Linf_scale",&Linfscale,"scalar","scale the Linf reference line");
    parse_args.Add("-fig",&fig_dir,"dir","base figure directory");

    parse_args.Parse();

    std::string boundary_file_tex=boundary_file+".tex";
    std::string boundary_file_text=boundary_file+".txt";
    Dump_Surface_data(boundary_file_text);
    Dump_Reference_data(L2scale,Linfscale);
    LOG::Finish_Logging();
    return 0;
}
